"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

"""
This file sets up a client adhering to the interface specified by the Qunicorn middleware
developed in the SeQuenC project: https://sequenc.de/

To learn more about Qunicorn check out the Qunicorn GitHub: https://github.com/qunicorn/qunicorn-core
And it's documentation:
https://qunicorn-core.readthedocs.io/en/latest/index.html


"""
import requests
import time

class BackendClient:
    """
    This object allows connecting to Qunicorn backend servers.

    Parameters
    ----------
    socket_ip : string
        The IP address of the socket of the target server.
    port : int
        The port on which the server is listening.


    Examples
    --------

    We assume that the example from BackendServer has been executed in the same console.

    >>> from qrisp.interface import BackendClient
    >>> example_backend = BackendClient(api_endpoint = "127.0.0.1", port = 8080)
    >>> from qrisp import QuantumCircuit
    >>> qc = QuantumCircuit(2)
    >>> qc.h(0)
    >>> qc.cx(0,1)
    >>> qc.measure(qc.qubits)
    >>> example_backend.run(qc, shots = 1000)
    {'00': 510, '11': 490}

    """
    
    def __init__(self, api_endpoint, port=None, token = ""):
        
        #https anstatt http
        api_endpoint = 'http://' + api_endpoint
        #if api_endpoint[:8] != 'http://':
        #    api_endpoint = 'http://' + api_endpoint

        
        
        
        if port is None:
            port = 9010
        self.port = port
        self.token = token
        
        self.api_endpoint = api_endpoint + ":" + str(port)
        
    #Executes 
    def run(self, qc, shots):
        qasm_str = qc.qasm()
        
        deployment_data = {
                        "programs": [
                                {
                                    "quantumCircuit": qasm_str,
                                    "assemblerLanguage": "QASM2",
                                    "pythonFilePath": "",
                                    "pythonFileMetadata": ""
                                    }
                                    ],
                                "name": ""
                            }
        deployment_response = requests.post(f'{self.api_endpoint}/deployments', json = deployment_data, verify = False)

        if deployment_response.status_code == 422:
            raise Exception(f'Unprocessable quantum ciruict {deployment_response.status_code}')
        elif deployment_response.status_code != 201:
            print(deployment_response.status_code)
            raise Exception(f'Failed to deploy quantum circuit {deployment_response.status_code}')
        
        deployment_id = deployment_response.json()["id"]
        
        job_data = {
              "name": "",
              "providerName": "",
              "deviceName": "",
              "shots": shots,
              "token": self.token,
              "type": "RUNNER",
              "deploymentId": deployment_id
            }
        
        job_post_response = requests.post(f'{self.api_endpoint}/jobs', json = job_data, verify = False)
        if deployment_response.status_code == 422:
            raise Exception(f'Unprocessable quantum ciruict (status code: {job_post_response.status_code})')
        elif deployment_response.status_code != 201:
            raise Exception(f'Failed to post job (status code: {job_post_response.status_code})')
        
        job_id = job_post_response.json()["id"]
        
        job_running = True
        

        while True:
        
            job_get_response = requests.get(f'{self.api_endpoint}/jobs/{job_id}', json = job_data, verify = False)
            
            if job_get_response.status_code != 201:
                raise Exception(f'Quantum circuit execution failed: {job_get_response.json()["message"]}')
            
            job_state = job_get_response.json()["state"]
        
            if job_state == "finished":
                break
                
            time.sleep(0.1)
            
        
        results = job_get_response.json()["results"][0]["results"]
        
        return results



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# This file implements the algorithm presented here:
# https://arxiv.org/abs/2304.02921

# We begin with some functions for carry propagation.
# For this we implement the Brent-Kung-Tree in a recursive manner. See:
# https://web.stanford.edu/class/archive/ee/ee371/ee371.1066/lectures/lect_04.pdf Page 13

from qrisp import *

# Returns the PROPAGATE status of a group of entries
def calc_P_group(P):
    new_p = QuantumBool(name = "p_group*")
    if len(P) == 2:
        mcx(P, new_p, method = "gidney")
    else:
        mcx(P, new_p)
    return new_p
    

# Returns the GENERATE status of a group of entries
# A group has GENERATE status True, if any of the members
# has GENERATE status True and all of the following members
# have PROPAGATE status True.

# The GENERATE status of the group is calculated into the last
# member of the group (G[-1])
def calc_G_group(P, G):
    for i in range(len(G)-1):
        mcx([G[i]] + P[i+1:], G[-1], method = "jones")
    return G[-1]
counter = 0
# Receives the values PROPAGATE and GENERATE values of a group.
# It is assumed that the 0-th entry of G already contains the carry status
# This function then calculates the carry status for every other entry of G
def propagate_carry(P, G):
    
    # This loop iteratively calculates the CARRY status into all entries
    # of G.
    
    # We calculate the CARRY status of position i by considering the CARRY status of
    # position 0 and the GENERATE status of any position above that. To successfully
    # generate a CARRY status of True, all the other positions in between need to have
    # PROPAGATE status True.
    
    # Note that in order to calculate the CARRY status of position i, we need the
    # GENERATE status (not the CARRY status) of j for all j < i.
    # This is why we reverse the loop, such that the position with the highest i
    # are calculated first (the computation can access all the GENERATE 
    # information because none of the CARRY status has been calculated yet)
    for i in range(1, len(G))[::-1]:
        for j in range(i):
            mcx([G[j]] + P[j+1:i+1], G[i], method = "jones")
        

# This function recursively calculates the layers of the Brent-Kung tree.
# It receives the GENERATE and CARRY status of a set of positions and returns
# the evaluated CARRY status

# The function takes the parameters P and G, which represent the PROPAGATE and
# the GENERATE status of the corresponding digit.

# The parameter r describes the radix of the Brent-Kung tree.

# Finally the parameter k_out describes a the layer where the carry calculation
# should be ended prematurely. That is if k_out = 2 and r = 4, the output will
# be every 4**2 = 16-th bit of G, that contains a properly calculated carry value
    

def brent_kung_tree(P, G, r, k_out = np.inf):
    # if len(G):
        # print(G[0].qs().compile(40, compile_mcm = True, gate_speed = gate_speed).t_depth())
    if r <= 1 or not isinstance(r, int):
        raise Exception("Provided invalid radix base (needs to be integer > 1)")
    
    # We group the positions and calculate the GENERATE and PROPAGATE status
    # of the groups. Subsequently the function is called recursively to evaluate the
    # CARRY status of the respective groups.
    
    # Finally the CARRY status of the groups is propagated to the remaining positions
    # to yield the complete set of carries (not only the groups)    
    
    # Recursion cancellation condition
    if len(G) == 0:
        return []

    # Set up list for grouped GENERATE and PROPAGATE states
    grouped_P = []
    grouped_G = []
    for i in range(len(G)//r):
        interval = slice(r*i,r*(i+1))
        
        if i == 0:
            # We can skip the 0-th position
            # This is because the propagate_carry function and the calc_G_group
            # function are not taking the 0-th position into consideration
            grouped_P.append(None)
        else:
            grouped_P.append(calc_P_group(P[interval]))
            
        grouped_G.append(calc_G_group(P[interval], G[interval]))
    
    # Call the function recursively
    grouped_carry = brent_kung_tree(grouped_P, grouped_G, r, k_out = k_out - 1)
    
    if k_out > 0:
        return grouped_carry
    
    # If working with a radix > 2, the carry also needs to be propagated
    # to the position of the first group. In the case of r = 2, this is not neccessary
    # because the first group consists just of the 0-th position (which already contains 
    # the carry in its GENERATE status) and the 1st position which contains the CARRY status
    # due to the GENERATE computation of this group.
    initial_interval = slice(0, r-1)
    propagate_carry(P[initial_interval], G[initial_interval])

    # Go through all the groups and propagate the carry.
    #Note that the last interval has not neccessarily size r (depending on the size of G)
    for i in range(1, len(G)//r+1):
        interval = slice(r*i-1, r*(i+1)-1)
        propagate_carry(P[interval], G[interval])
    
    # Return G (now contains the carry positions)
    return G


# The following function calculates the carry status from two QuantumVariables of equal size

# The function takes two QuantumFloats a, b, the radix base and the radix exponent.
# The function returns a QuantumVariable c, that contains the CARRY status of every
# radix_base**radix_exponent digit of CARRY(a,b)

# Example:
    
# a = 3 = 0011
# b = 6 = 0110
# CARRY(a,b) = 0110

# Assume we choose radix_base = 2. If we choose radix_exponent = 0,
# this function returns

# c = 0110

#If we choose radix_exponent = 1

# c = 01

# ie. position 1 and 3 of CARRY(a,b)

# If we chose radix_exponent = 2

# c = 0

# ie. position 3 of CARRY(a,b)


# Furthermore this function uncomputes all garbage. What is the garbage here?
# The cancelation of the brent kung tree at an early layer produce many GENERATE
# entries, that are not holding CARRY values. Furthermore, also the PROPAGATE values
# of groups are uncomputed.

@auto_uncompute
def calc_carry(a, b, radix_base = 2, radix_exponent = 0):
    
    R = radix_base**radix_exponent
    # How can we achieve that the GENERATE entries, that don't contain a relevant
    # CARRY information are uncomputed while the CARRY entries stay?
    
    # We do this by creating two QuantumVariables, one which holds the carry and
    # one which holds the GENERATE garbage. We then arrange both of them in a list g,
    # that represents the GENERATE array.
    
    # This variable will hold the result
    # If b can be divided into k blocks of size R,
    # we only need k-1 ancillae qubit, because we have no need for
    # the carry of the last bock.
    c = QuantumVariable(int(np.ceil(len(b)/R))-1)
    
    # This variable will hold the intermediate GENERATE values, that are supposed 
    # to be uncomputed. The uncomputation is performed using the auto_uncompute 
    # decorator. This decorator uncomputes all local variables.
    brent_kung_ancilla = QuantumVariable(c.size*(R-1))
    
    #Create the g list
    anc_list = list(brent_kung_ancilla)
    c_list = list(c)
    g = []
    for i in range(len(c)):
        for j in range(R-1):
            g.append(anc_list.pop(0))
        g.append(c_list.pop(0))
    
    # Append the remaining ancillae
    g.extend(anc_list)
    
    # We now compute the GENERATE and the PROPAGATE status of each digit
    # g_i = a_i * b_i
    # p_i = a_i XOR b_i
    
    # This could be achieved by applying a sequence of mcx gates to compute g
    # followed by a sequence of cx gates that compute p.
    
    # This however breaks the automatic uncomputation algorithm since it induces
    # a cycle in the dag representation. We can still use automatic uncomputation
    # with a hack: We use both operation together and mark the argument of b as
    # permeable.
    
    # This gate is synthesized by the anc_gate function.
    
    for i in range(min(len(g), len(a), len(b))):
        a.qs.append(anc_gate(method = "gidney"), [a[i], b[i], g[i]])
        
    # In some rare cases this hack seems to also disturb the gate ordering, since
    # the dag linearization assumes commutativity due to the permeability. We fix this problem
    # by enforcing the order with a non-permeable identity. Ie. two subseqeuent x gates
    x(g)
    x(g)
    # These x gates are cancelled by the compile method so there is no overhead.
    
    p = b
    
    # Calculate compute the carry using the Brent-Kung tree
    brent_kung_tree(p, g, radix_base, radix_exponent)
    
    # print("======")
    # print(c.qs.compile(40, compile_mcm = True, gate_speed = gate_speed).t_depth())

    # Undo the previous cx to keep b unchanged
    # For the entries that are uncomputed automatically, the inverted anc_gate 
    # will restore b So we only need to restore.
    for i in range(min(len(g), len(a), len(b))):
        if g[i] in list(c):
            cx(a[i], b[i])
        else:
            pass

    # Return the result
    return c

# This function returns a gate, that performs the toffoli and cx gate. 
# We mark the second qubit as permeable to "hack" the uncomputation algorithm.
def anc_gate(method = "gray"):
    anc_qc = QuantumCircuit(3)
    anc_qc.mcx([0,1], 2, method = method)
    anc_qc.cx(0,1)
    res = anc_qc.to_gate(name = "anc_gate")
    res.permeability = {0 : True, 1: True, 2: False}
    res.is_qfree = True
    return res
    
# This function performs the in-place addition 
# b += a
# based on the higher radix qcla
# The overall radix can be specified as an exponential of the form
# R = radix_base**radix_exponent
def qcla(a, b, radix_base = 2, radix_exponent = 0, reduce_t_depth = True):
    
    if len(a) > len(b):
        raise Exception("Tried to add QuantumFloat of higher precision onto QuantumFloat of lower precision")
    
    R = radix_base**radix_exponent
    
    # The case that a only has a single qubit is simple.
    if len(b) == 1:
        cx(a,b)
        return
    elif len(b) <= R:
        qcla_anc = QuantumVariable(len(b) - len(a))
        gidney_adder(list(a) + list(qcla_anc), b)
        qcla_anc.delete()
        return
    
    
    
    # Calculate the carry
    c = calc_carry(a, b, radix_base, radix_exponent)
    
    sum_path_gidney(a,b,c,R)

    
    
def sum_path(a, b, c, R):
    
    if R == 1:
        # If R = 1, we are in the case of Drapers QCLA:
        # https://arxiv.org/abs/quant-ph/0406142   
        # We can use the formula
        # S = A (+) B (+) C
        cx(c[:-1], b[1:])
        cx(a, b[:len(a)])
    else:
        i = 0
        # Execute addition using the corresponding carry values
        for i in range(len(c)+1):
            
            # Determine the radix qubits to perform the addition on
            a_block = a[R*i:R*(i+1)]
            b_block = b[R*i:R*(i+1)]
            
            # If a_block is empty we break the loop.
            # This doesn't necessarily imply that b_block is empty because
            # len(b) could be larger then len(a). In this case we execute an
            # increment block after this loop
            if not len(a_block):
                i = i-1
                break
            
            # If a_block and b_block have different sizes, we create a padding
            # variable to fill up the remaining entries
            if not len(a_block) == len(b_block):
                padding_var = QuantumVariable(len(b_block) - len(a_block))
                a_block = a_block + list(padding_var)
            # Perform Cuccarro addition
            if i == 0:
                # cuccaro_procedure(a.qs, a_block, b_block)
                
                gidney_adder(a_block, b_block)
            else:
                
                # cuccaro_procedure(a.qs, a_block, b_block, carry_in = c[i-1])
                gidney_adder(a_block, b_block, c[i-1])
            
            # Delete carry var
            try:
                padding_var.delete()
            except NameError:
                pass
        
        
        # This loop treats the case that len(a) != len(b)
        # In this case we need to perform the increment function on the remaining
        # qubits of b (incase the carry is True)
        #We start at the index the last loop finished at
        for j in range(i+1, len(b)//R+1)[::-1]:
            b_block = b[R*j:R*(j+1)]
            
            
            if not len(b_block):
                continue
            
            #Perform incrementation function
            if R*j == 0:
                qbl = QuantumBool()
                lin_incr(b_block, qbl[0], c_out = c[j])
                qbl.delete()
            else:
                lin_incr(b_block, c[j-1], c_out = c[j])
            
    
    # To uncompute the carry we use Drapers strategy
    # CARRY(A,B) = CARRY(A, NOT(A+B))
    # We therefore bitflip the sum
    x(b)
    
    # Contrary to Draper's adder we don't need to uncompute every carry digit.
    # Because of the above equivalence, the carries agree on every digit, so especially
    # on the digits representing the output of the calc_carry function. We can therefore
    # uncompute using calc_carry (even with higher radix) by inverting calc_carry.
    
    with invert():
        #We use the redirect_qfunction decorator to steer the function onto c
        redirect_qfunction(calc_carry)(a, b, radix_base, radix_exponent, target = c)
    
    # Delete c. If needed, verify that the uncomputation indeed suceeded
    c.delete(verify = False)
    
    # Flip the sum back
    x(b)
    

def sum_path_gidney(a, b, c, R):
    
    if R == 1:
        # If R = 1, we are in the case of Drapers QCLA:
        # https://arxiv.org/abs/quant-ph/0406142   
        # We can use the formula
        # S = A (+) B (+) C
        cx(c, b[1:])
        
        # We now do the mother of all evil hacks:
        # We can get rid of c using automatic uncomputation.
        # However since we already hacked the uncomputation by claiming
        # the anc_gate is permeable when in reality it's not, the anc_gate
        # (which is used to uncompute c) introduces an additional cx gate.
        # This cx gate incidentally cancels out with the one from below.
        
        c.uncompute()
        # cx(a, b[:len(a)])
        return
    
    # print(c)
    # # This loop treats the case that len(a) != len(b)
    # # In this case we need to perform the increment function on the remaining
    # # qubits of b (incase the carry is True)
    # print(c)
    # #We start at the index the last loop finished at
    if len(a) != len(b):
        # print("len(a): ", len(a))
        # print("R: ", R)
        # print("len(a)//R: ", len(a)//R)
        for j in range(int(np.ceil(len(a)/R)), len(b)//R+2)[::-1]:
            
            b_block = b[R*j:R*(j+1)]
            # print(j)
            
            if not len(b_block):
                continue
            # print(b_block)
            
            #Perform incrementation function
            if R*j == 0:
                lin_incr(b_block, c_out = c[j])
            else:
                if j < len(c):
                    lin_incr(b_block, c[j-1], c_out = c[j])
                else:
                    lin_incr(b_block, c[j-1])
                    

    # Execute addition using the corresponding carry values
    for i in range(len(a)//R+1)[::-1]:
        # Determine the radix qubits to perform the addition on
        a_block = a[R*i:R*(i+1)]
        b_block = b[R*i:R*(i+1)]
        
        # If a_block is empty we break the loop.
        # This doesn't necessarily imply that b_block is empty because
        # len(b) could be larger then len(a). In this case we execute an
        # increment block after this loop
        if not len(a_block):
            i = i-1
            continue
        
        
        # If a_block and b_block have different sizes, we create a padding
        # variable to fill up the remaining entries
        if not len(a_block) == len(b_block):
            padding_var = QuantumVariable(len(b_block) - len(a_block))
            a_block = a_block + list(padding_var)
        
        
        # Perform Cuccarro addition
        if i == 0:
            # cuccaro_procedure(a.qs, a_block, b_block)
            if len(c):
                gidney_adder(a_block, b_block, c_out = c[i])
            else:
                gidney_adder(a_block, b_block)
        elif i < len(c):
            # cuccaro_procedure(a.qs, a_block, b_block, carry_in = c[i-1])
            gidney_adder(a_block, b_block, c_in = c[i-1], c_out = c[i])
        else:
            gidney_adder(a_block, b_block, c_in = c[i-1])
            
        
        # Delete carry var
        try:
            padding_var.delete()
        except NameError:
            pass
        
    
    c.delete(verify = False)
    
    
    
        

# This function performs the linear depth incrementation described in 
# https://algassert.com/circuits/2015/06/12/Constructing-Large-Increment-Gates.html
# This function also needs a linear amount of qubits. These qubits pose no overhead
# because in the above QCLA function, the calc_carry function already allocated and
# deallocated some ancillas, so the compiler can reuse these
def lin_incr(a, c_in = None, c_out = None):
    
    if c_out is not None:
        a = list(a) + [c_out]
    
    if len(a) == 1:
        cx(c_in,a)
        return
    
    incr_anc = QuantumVariable(len(a)-1)
    
    if c_in is not None:
        mcx([c_in, a[0]], incr_anc[0], method = "gidney")
    
    for i in range(len(a)-2):
        mcx([incr_anc[i], a[i+1]], incr_anc[i+1], method = "gidney")
        
    cx(incr_anc[-1], a[-1])
    
    for i in range(len(a)-2)[::-1]:
        mcx([incr_anc[i], a[i+1]], incr_anc[i+1], method = "gidney_inv")
        cx(incr_anc[i], a[i+1])
    
    if c_in is not None:
        mcx([c_in, a[0]], incr_anc[0], method = "gidney_inv")
        cx(c_in, a[0])
        
    incr_anc.delete()
    


a = QuantumFloat(4)
b = QuantumFloat(4)
d = b.duplicate()
# a = QuantumVariable(3)
# b = QuantumVariable(3)

# c = QuantumBool()
a[:] = 1
b[:] = 0
# c[:] = 1
# sum_path_gidney(a, b, c, 2)

# h(a)
# h(b)
# d[:] = b

# print(multi_measurement([a,b,c]))

# import time


radix_base = 2
radix_exponent = 1
c = calc_carry(a, b, radix_base, radix_exponent)
# # brent_kung_tree(a, b, radix_b1ase, radix_exponent)
# # print(counter)
# qcla(a, b, radix_base, radix_exponent)

# print(a.qs)
# b += a
# gidney_adder(a, b)

print(c)
from qrisp.interface import QiskitBackend
# print(multi_measurement([a,d,b]))
# print(b.get_measurement(compilation_kwargs = {"compile_mcm" : True}))
# print(b)

# print(c)
# print(a.qs.depth())
# gate_speed = lambda x : t_depth_indicator(x, 2**-6)
# print(a.qs.compile(40, compile_mcm = True, gate_speed = gate_speed).t_depth())
# print(a.qs.compile(2).cnot_count())
# print(a.qs.compile(2).num_qubits())
# print(a.qs.count_ops())
#%%
# Test functions
# Function to classically calculate the carry value
def full_adder_carry(bitstring1, bitstring2):
    if len(bitstring1) != len(bitstring2):
        raise ValueError("Bitstrings must be of equal length")

    n = len(bitstring1)
    carry = 0
    carry_values = []

    for i in range(n):
        a = int(bitstring1[i])
        b = int(bitstring2[i])

        # Calculate the carry-out (Cout) for this bit
        sum_ab = a + b + carry
        carry_out = sum_ab // 2

        carry_values.append(carry_out)

        # Update the carry for the next iteration
        carry = carry_out

    return ''.join(str(c) for c in carry_values)

    
n = 4
for n in range(3, 8):
    for r in range(2,n):
        a = QuantumVariable(n)
        b = QuantumVariable(n)
        
        h(a)
        h(b)
                
        c = calc_carry(a, b, r)
        
        mes_res = multi_measurement([a, b, c])
        
        for k in mes_res.keys():
            if k[2] != full_adder_carry(k[0], k[1])[:-1]:
                print(k[0])
                print(k[1])
                print(k[2])
                print(full_adder_carry(k[0], k[1])[:-1])
                
                assert False

#%%
# for R in [1,2,4,8]:
    
import time

t0 = time.time()
for radix_base in [2,3]:
    for radix_exponent in [1,0,2]:
        for m in range(1, 7):        
            for n in range(1, m):
                
                a = QuantumFloat(n)
                b = QuantumFloat(m)
                c = QuantumFloat(m)
                
                h(a)
                h(b)
                c[:] = b
                
                print(a.size, b.size, radix_base, radix_exponent)
                qcla(a, b, radix_base = radix_base, radix_exponent = radix_exponent)
                mes_res = multi_measurement([a, c, b])
                
                for k in mes_res.keys():
                    if k[2] != (k[0] + k[1])%2**m:
                        print(k[0])
                        print(k[1])
                        print(k[2])
                        print((k[0]+k[1])%2**m)
                        
                        assert False
                
                statevector_arr = a.qs.compile().statevector_array()
                angles = np.angle(
                    statevector_arr[
                        np.abs(statevector_arr) > 1 / 2 ** ((a.size + b.size) / 2 + 1)
                    ]
                )
    
                # Test correct phase behavior
                assert np.sum(np.abs(angles)) < 0.1
                
                
print(time.time()-t0)


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import p, cp, cx, mcp, QuantumVariable
import numpy as np

class QuantumColor(QuantumVariable):
    """
    The QuantumColor is a custom QuantumVariable implemented with tackling the Max-k-Colorable-Subgraph problem
    and other coloring optimization problems in mind. It provides flexibility in choosing encoding methods and 
    leverages efficient data structures like QuantumArrays to enhance computational performance.

    The QuantumColor class takes as input a list of colors and a flag indicating the preferred encoding 
    method - binary or one-hot encoding. The choice of encoding method has implications for how colors are 
    represented in the quantum computation.

    In binary encoding, each color is represented by a unique binary number. For instance, if there are four 
    colors, Red, Green, Blue, and Yellow, they can be represented, for example, as [0,0], [0,1], [1,0], and [1,1] 
    respectively.

    In contrast, one-hot encoding represents each color as an array where only one element is 1 and the rest are 0. 
    Using the same four-color example, red can be represented as [1,0,0,0], green as [0,1,0,0], blue as [0,0,1,0], 
    and yellow as [0,0,0,1].

    Another key feature of the QuantumColor class is its use of QuantumArrays. QuantumArrays are data structures designed for efficient quantum computation. 
    They allow for compact representation and manipulation of quantum states and operators.

    Parameters.
    ----------
    list_of_colors : list
        The list of colors to be used in the quantum coloring problem.
    one_hot_enc : bool, optional
        The flag to indicate whether to use one-hot encoding. If False, binary encoding is used. We use the one-hot encoding by default.

    Attributes
    ----------
    list_of_colors : list
        The list of colors to be used in the quantum coloring problem.
    one_hot_enc : bool
        The indicator which tells the program whether to use one-hot encoding. If False, binary encoding is used.

    Methods
    -------
    decoder(i)
        Decode the color from the given index for both binary and one-hot encoding.
    """

    def __init__(self, list_of_colors, one_hot_enc = True): 
        """
        Initialize the QuantumColor with a list of colors and a flag indicating whether to use one-hot encoding.

        Parameters
        ----------
        list_of_colors : list
            The list of colors to be used in the coloring problem instance.
        one_hot_enc : bool, optional
            The flag to indicate whether to use one-hot encoding. If False, binary encoding is used. Default is True.

        """
        self.list_of_colors = list_of_colors
        self.one_hot_enc = one_hot_enc

        # If one-hot encoding is used, the size of QuantumVariable is the number of colors
        if one_hot_enc:
            QuantumVariable.__init__(self, size = len(list_of_colors)) 

        # If binary encoding is used, the size of QuantumVariable is the maximal value of log2 for the number of colors
        else:
            QuantumVariable.__init__(self, size = int(np.ceil(np.log2(len(list_of_colors)))))

    def decoder(self, i):
        """
        Decode the color from the given index i.

        Parameters
        ----------
        i : int
            The index to be decoded into a color.

        Returns
        -------
        str
            The decoded color if it exists, otherwise "undefined".
        
        """
        if not self.one_hot_enc:
            # Binary encoding: Each color is represented by a binary number.

            # For example, with four colors Red, Green, Blue and Yellow:
            #Red:   [0,0]
            #Green: [0,1]
            #Green: [1,0]
            #Yellow:[1,1]
            return self.list_of_colors[i]

        else:
            #One hot encoding: Each color is represented by an array where only one element is 1 and rest are 0.

            # For example, with four colors Red, Green, Blue and Yellow:     
            #Red:   [1,0,0,0]
            #Green: [0,1,0,0]
            #Yellow:[0,0,1,0]
            #Blue:  [0,0,0,1]

            is_power_of_two = ((i & (i-1) == 0) and i != 0)

            if is_power_of_two:
                return self.list_of_colors[int(np.log2(i))]

            else:
                return "undefined"
            
def initial_state_mkcs(qarg):
    """
    The initial_state_mkcs function provides the correct initial state of qubits in
    the system on which we run the optimization. In the case of the Max-k-Colorable 
    Subgraph problem, the initial state of the systemis simply any random coloring 
    of nodes of the graph.

    Parameters
    ----------
    qarg : QuantumArray
        A QuantumArray consisting of the color values for each node of graph G.


    Returns
    -------
    qarg : QuantumArray
        The quantum argument (in our case this is a QuantumArray) adapted to include 
        the information of the initial state of the system.
          
    """

    # Set all elements in qarg to initial state
    qarg[:] = init_state

    # Return updated quantum argument
    return qarg

def apply_phase_if_eq(qcolor_0, qcolor_1, gamma):
    """
    Applies a phase if the colors of the two arguments are matching.

    Parameters
    ----------
    qcolor_0 : QuantumColor
        The color of the first argument
    qcolor_1 : QuantumColor
        The color of the second argument
    gamma : float
        The value of the gamma angle parameter used in QAOA

    Returns
    -------
    None.
        Applies a phase if the colors of the two arguments are matching.

    """
    if qcolor_0.one_hot_enc !=  qcolor_1.one_hot_enc:
          raise Exception("....")
    #Raise exception if color list if different
    if qcolor_0.one_hot_enc:
        #qbl = (qcolor_0 != qcolor_1)
        #p(2 * gamma, qbl)
        #qbl.uncompute()
        for i in range(qcolor_0.size):
            cp(2*gamma, qcolor_0[i], qcolor_1[i])

    else:
        cx(qcolor_0, qcolor_1)
        mcp(2*gamma, qcolor_1, ctrl_state = 0)
        cx(qcolor_0, qcolor_1)

def create_coloring_operator(G):
    r"""
    Generates the cost operator for an instance of the coloring problem for a given graph ``G``.

    The coloring operator and applies a phase if two neighboring nodes in the graph have the same color.

    Parameters
    ----------
    G : nx.Graph
        The graph to color.

    Returns
    -------
    cost_operator : function
        A function receiving a :ref:`QuantumVariable` and a real parameter $\gamma$.
        This function performs the application of the cost operator.

    """
    def coloring_operator(quantumcolor_array, gamma):
        for pair in list(G.edges()):
            apply_phase_if_eq(quantumcolor_array[pair[0]],
                              quantumcolor_array[pair[1]], 
                              gamma)
    return coloring_operator 

def mkcs_obj(quantumcolor_array, G):
        # Set value of color integer to 1
    cost = 1

        # Iterate over all edges in graph G
    for pair in list(G.edges()):

            # If colors of nodes in current pair are not same, multiply color by reward factor 4
        if quantumcolor_array[pair[0]] != quantumcolor_array[pair[1]]:
            cost *= 4

        # Return negative color as objective function value. The negative value is used since we want to minimize the objective function       
    return -cost

def create_coloring_cl_cost_function(G):
    """
    Creates the classical cost function for an instance of the coloring problem for a given graph ``G``.

    Parameters
    ----------
    G : nx.Graph
        The graph to color.

    Returns
    -------
    cl_cost_function : function
        The classical cost function for the problem instance, which takes a dictionary of measurement results as input.

    """
    def cl_cost_function(res_dic):

        def mkcs_obj(quantumcolor_array, G):
            cost = 1
            for pair in list(G.edges()):
                if quantumcolor_array[pair[0]] != quantumcolor_array[pair[1]]:
                    cost *= 4
            return -cost
    
        cost = 0
        for quantumcolor_array, prob in res_dic.items():
            cost += mkcs_obj(quantumcolor_array,G)*prob

        return cost
    
    return cl_cost_function


def graph_coloring_problem(G):
    """
    Creates a QAOA problem instance with appropriate phase separator, mixer, and
    classical cost function.

    Parameters
    ----------
    G : nx.Graph
        The graph to color.

    Returns
    -------
    QAOAProblem : function
        A QAOA problem instance for graph coloring for a given graph ``G``.

    """    
    from qrisp.qaoa import QAOAProblem, apply_XY_mixer

    return QAOAProblem(create_coloring_operator(G), apply_XY_mixer, create_coloring_cl_cost_function(G))


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# -*- coding: utf-8 -*-

# This file contains some functions that enhance the speed
# of SparseBiArray manipulation

import numpy as np
from numba import uint64, uint32, int64, int32, njit, prange, vectorize
from qrisp.simulator.numerics_config import float_tresh, cutoff_ratio
from scipy.sparse import coo_array

# It can happen that the coo-matrix multiplication puts two data entries
# for the same index. These function recognizes and sums these duplicates
def sum_duplicates(nz_indices, data):
    new_indices, reconstr_indices = np.unique(nz_indices, return_inverse=True)

    if len(nz_indices) > 1e4:
        new_data = sum_duplicates_aux_jitted(new_indices, data, reconstr_indices)
    else:
        new_data = sum_duplicates_aux(new_indices, data, reconstr_indices)

    return new_indices, new_data


def sum_duplicates_aux(new_indices, old_data, reconstr_indices):
    new_data = np.zeros(new_indices.shape, dtype=old_data.dtype)

    for i in range(len(reconstr_indices)):
        new_data[reconstr_indices[i]] += old_data[i]

    return new_data


@njit(nogil=True, cache=True)
def sum_duplicates_aux_jitted(new_indices, old_data, reconstr_indices):
    new_data = np.zeros(new_indices.shape, dtype=old_data.dtype)

    for i in range(len(reconstr_indices)):
        new_data[reconstr_indices[i]] += old_data[i]

    return new_data


# This function rounds entries, which are below a certain float thresh and removes the
# tracked indices.
# @njit(parallel = True)
def elim_zeros(nz_indices, data, decimals=8):
    # return nz_indices, data
    # data = np.around(data, decimals=decimals, out=data)

    # mask = np.abs(data) > 10**-decimals
    mask = np.nonzero(data)[0]

    return nz_indices[mask], data[mask]


# This function applies an index permutation onto
# flattened indices.
# Consider the situation that we have the two flattened indices
# [1,3]
# of a 2x2 matrix.
# If we now want to permute the axes, we permute the bit level
# representation of these numbers
# ["01", "10"] ==> ["10", "01"]


@njit(nogil=True, cache=True)
def permute_axes(index, permutation):
    permutation = -permutation + len(permutation) - 1
    permutation = permutation[::-1]
    new_index = np.zeros(index.size, dtype=index.dtype)
    for i in range(len(permutation)):
        if permutation[i] == i:
            # new_index = new_index | (index&(1<<permutation[i]))
            for j in prange(len(new_index)):
                new_index[j] = new_index[j] | (index[j] & (1 << permutation[i]))
        else:
            # new_index = new_index |
            # (((index&(1<<permutation[i]))>>permutation[i])<<(i))
            for j in prange(len(new_index)):
                new_index[j] = new_index[j] | (
                    ((index[j] & (1 << permutation[i])) >> permutation[i]) << (i)
                )

    return new_index


# This function retrieves the coordinates from
# flattened coordinades for a given shape of length two
def get_coordinates(indices, shape):
    if shape[0] < 1<<31:
        coordinates_0 = get_coordinates_0_32(indices, int(np.log2(shape[1])))
    else:
        coordinates_0 = get_coordinates_0_64(indices, int(np.log2(shape[1])))
        
    if shape[1] < 1<<31:
        coordinates_1 = get_coordinates_1_32(indices, int(np.log2(shape[1])))
    else:
        coordinates_1 = get_coordinates_1_64(indices, int(np.log2(shape[1])))

    return coordinates_0, coordinates_1
        


@vectorize([uint64(int64, int64)])
def get_coordinates_0_64(index, bit_shape_1):
    return index >> bit_shape_1


@vectorize([uint64(int64, int64)])
def get_coordinates_1_64(index, bit_shape_0):
    return index & ((1 << bit_shape_0) - 1)

@vectorize([uint32(int64, int64)])
def get_coordinates_0_32(index, bit_shape_1):
    return index >> bit_shape_1

@vectorize([uint32(int64, int64)])
def get_coordinates_1_32(index, bit_shape_0):
    return index & ((1 << bit_shape_0) - 1)



# This function basically inverts get_coordinates.
# Ie. it generates flat coordinates from row and column indices
def gen_flat_coords(col, row, shape):
    return gen_flat_coords_vec(col, row, int(np.log2(shape[1])))
    
@vectorize([int64(int64, int64, int64)], cache=True)
def gen_flat_coords_vec(col, row, bit_shape_1):
    return row << bit_shape_1 | col



# This function constructs a flattened numpy array from
# given flattened coordinates and data
@njit(nogil=True, cache=True)
def construct_flat_array(flat_coords, data, size):
    res = np.zeros(size, dtype=data.dtype)

    for i in range(len(flat_coords)):
        res[flat_coords[i]] += data[i]

    return res

def construct_flat_array_no_jit(flat_coords, data, size):
    res = np.zeros(size, dtype=data.dtype)

    for i in range(len(flat_coords)):
        res[flat_coords[i]] += data[i]

    return res


@njit(cache=True)
def extract_bit_range(index, start, stop):
    breadth = stop - start
    return (index & (((1 << breadth) - 1) << start)) >> start


@njit(cache=True)
def insert_bit_range(target, bit_string, start):
    target |= bit_string << start


@njit(nogil=True, cache=True)
def jitted_permuter(index, new_index, bit_partition, permuted_bit_partition, perm):
    # for i in range(len(perm)):
    #     # bit_range = extract_bit_range(index, bit_partition[i], bit_partition[i+1])
    #     # insert_bit_range(new_index, bit_range, permuted_bit_partition[perm[i]])
    #     #Inlined version more efficient:
    #     new_index |= (((index&((((1<<(bit_partition[i+1] -
    #     bit_partition[i]))-1))<<bit_partition[i]))>>bit_partition[i])<<
    #     permuted_bit_partition[perm[i]])

    for j in prange(index.size):
        for i in range(len(perm)):
            # bit_range = extract_bit_range(index, bit_partition[i], bit_partition[i+1])
            # insert_bit_range(new_index, bit_range, permuted_bit_partition[perm[i]])
            # Inlined version more efficient:
            new_index[j] |= (
                (
                    index[j]
                    & (
                        ((1 << (bit_partition[i + 1] - bit_partition[i])) - 1)
                        << bit_partition[i]
                    )
                )
                >> bit_partition[i]
            ) << permuted_bit_partition[perm[i]]


def permuter(index, new_index, bit_partition, permuted_bit_partition, perm):
    for i in range(len(perm)):
        # bit_range = extract_bit_range(index, bit_partition[i], bit_partition[i+1])
        # insert_bit_range(new_index, bit_range, permuted_bit_partition[perm[i]])
        # Inlined version more efficient:
        if isinstance(new_index, np.ndarray):
            new_index |= (
                (
                    index
                    & (
                        (int(1 << (bit_partition[i + 1] - bit_partition[i])) - 1)
                        << bit_partition[i]
                    )
                )
                >> bit_partition[i]
            ) << permuted_bit_partition[perm[i]]
        else:
            for j in range(len(index)):
                new_index[j] |= (
                    (
                        index[j]
                        & (
                            (int(1 << (bit_partition[i + 1] - bit_partition[i])) - 1)
                            << int(bit_partition[i])
                        )
                    )
                    >> int(bit_partition[i])
                ) << int(permuted_bit_partition[perm[i]])
            


def invert_perm(perm):
    return [perm.index(i) for i in range(len(perm))]


def permute_axes(index, index_bit_permutation, jit = True):
    n = len(index_bit_permutation)

    index_bit_permutation = -np.array(index_bit_permutation) + n - 1
    index_bit_permutation = list(index_bit_permutation)[::-1]

    index_bit_permutation = list(index_bit_permutation)

    # Get the permutation of the axes

    perm = [index_bit_permutation.index(i) for i in range(n)]

    # Initialize the shape
    perm_shape = n * [1]

    i = 1
    while i < len(perm):
        # If this condition is met, both indices form an interval,
        # that does not need further permutation
        if perm[i] - 1 == perm[i - 1]:
            # Remove the index from the perm and perm_shape list
            perm.pop(i)
            perm_shape.pop(i)

            # Update the shape
            perm_shape[i - 1] += 1
            # Update the permutation indices
            for j in range(len(perm)):
                if perm[j] > perm[i - 1]:
                    perm[j] -= 1
        else:
            i += 1

    if isinstance(index, np.ndarray):
        new_index = np.zeros(index.size, dtype=index.dtype)
        dtype = index.dtype
    else:
        new_index = [0 for _ in range(len(index))]
        dtype = np.int64

    bit_partition = [sum(perm_shape[:i]) for i in range(len(perm_shape) + 1)]

    perm_shape_permuted = [perm_shape[i] for i in invert_perm(perm)]
    permuted_bit_partition = [
        sum(perm_shape_permuted[:i]) for i in range(len(perm_shape) + 1)
    ]

    if len(perm) * len(index) > 2**10 and jit:
        jitted_permuter(
            index,
            new_index,
            np.array(bit_partition),
            np.array(permuted_bit_partition),
            np.array(perm),
        )
    else:
        
        permuter(
            index,
            new_index,
            np.array(bit_partition, dtype = dtype),
            np.array(permuted_bit_partition, dtype = dtype),
            np.array(perm, dtype = dtype),
        )

    return new_index


@njit(cache=True)
def invert_permutation(p):
    """Return an array s with which np.array_equal(arr[p][s], arr) is True.
    The array_like argument p must be some permutation of 0, 1, ..., len(p)-1.
    """
    s = np.empty(len(p), dtype=np.int64)
    s[p] = np.arange(p.size)
    return s


@njit(nogil=True, cache=True)
def bi_array_moveaxis(data_array, index_perm, f_index_array):
    for i in range(len(index_perm)):
        shifted_value = 1 << index_perm[i]
        start_index = 1 << i

        for j in range(start_index):
            f_index_array[start_index ^ j] = shifted_value ^ f_index_array[j]

    # return f_index_array
    return data_array[f_index_array]


@njit(parallel = True, cache = True)
def dense_measurement_brute(input_array, mes_amount, outcome_index):
    
    n = int(np.log2(len(input_array)))
    mes_amount = int(mes_amount)
    reshaped_array = input_array.reshape((2**(mes_amount), 2**(n-mes_amount)))
    
    p_array = np.zeros(2**mes_amount, dtype = input_array.dtype)
    
    for i in prange(2**mes_amount):
        new_array = reshaped_array[i,:]
        p_array[i] = np.vdot(new_array, new_array)
    
    p_array = np.abs(p_array)
    max_p_array = np.max(p_array)
    
    indices = np.nonzero(p_array > max_p_array*cutoff_ratio)[0]
    
    return reshaped_array[indices,:], p_array[indices], indices

    new_arrays = []
    p_values = []
    outcome_indices = []
    
    for i in range(2**mes_amount):
        p = p_array[i]
        if p > max_p_array*cutoff_ratio:
            new_arrays.append(reshaped_array[i, :])
            p_values.append(p)
            outcome_indices.append(outcome_index+i)
    
    return new_arrays, p_values, outcome_indices


@njit(nogil=True, cache=True)
def dense_measurement_smart(input_array, mes_amount, outcome_index):
    
    p = np.abs(np.vdot(input_array, input_array) )
    
    if p < float_tresh:
        return [input_array], [p], [-1]
    
    if mes_amount == 0:
        return [input_array], [p], [outcome_index]
    
    N = input_array.shape[0]
    new_arrays = []
    p_list = []
    outcome_index_list = []
    
    a, b, c = dense_measurement_smart(input_array[:N//2], mes_amount - 1, outcome_index)
    
    if c[0] != -1:
        new_arrays.extend(a)
        p_list.extend(b)
        outcome_index_list.extend(c)
    
    a, b, c = dense_measurement_smart(input_array[N//2:], mes_amount - 1, outcome_index + 2**(mes_amount-1))
    
    if c[0] != -1:
        new_arrays.extend(a)
        p_list.extend(b)
        outcome_index_list.extend(c)
    
    if len(outcome_index_list):
        return new_arrays, p_list, outcome_index_list
    else:
        return new_arrays, p_list, [-1]


@njit(cache = True)
def sort_indices_jitted(row, col, data, shape_1):
    
    shifted_row = row << (int(np.log2(shape_1))+1)
    sorting_array = shifted_row ^ col
    
    array_sort = np.argsort(sorting_array, kind = "mergesort")
    
    new_row = row[array_sort]
    new_col = col[array_sort]
    new_data = data[array_sort]
    
    return new_row, new_col, new_data


def coo_sparse_matrix_mult_inner(A_row, A_col, A_data, B_row, B_col, B_data, A_shape, B_shape):
    """
    This function describes a novel sparse matrix multiplication algorithm operating
    on the COO format. The scipy and MKL implementation of sparse matrix multiplication
    operate on the CSR/CSC format, which requires a conversion step, that can
    be costly for very sparse matrices.
    
    Given are the inputs A, B in sparse format and we want to extract A*B in sparse format
    
    The steps in the algorithm presented here are the following
    
    1. For both inputs, sort the data according to the indices, prioritizing columns
    for A and rows for B
    
    For instance if we have a sparse matrix with
    
    col = [1,2,1,3]
    row = [4,3,3,1]
    data = [0.1, 0.2, 0.3, 0.4]
    
    The sorting with columns prioritized is
    
    col = [1, 1, 2, 3]
    row = [3, 4, 3, 1]
    data = [0.3, 0.1, 0.2, 0.4]
    
    2. Find the indices, where the indices of the prioritized array increases.
    We call this the "unique marker". For instance in the example above,
    the first unique marker is at index 2, because at this index, the col array
    changes from 1 to 2 and the next one is at index 3 because the col array changes
    from 2 to 3.
    
    Each of these interval now represent one column of the matrix. For instance 
    the first column is represented by 
    
    col = [1,1]
    row = [3,4]
    data = [0.3, 0.1]
    
    3. We now identified the columns for A and the rows for B.
    Matrix multiplication is essentially the scalar product of the columns
    of A with the rows of B.
    
    To compute the dot product of a given column with a given row, we need to find
    the index agreements. This is because only on these indices, there can be
    a non-zero contribution. 
    
    For example assume that we have
    
    Column A:
        
    col_a = [1,1]
    row_a = [3,4]
    data_a = [0.3, 0.1]

    Row B:
        
    col_b = [1,4]
    row_b = [3,3]
    data_b = [0.7, 0.8]
    
    This computes the scalar product of column 1 of A with row 3 of B (therefore
    the result will be the (1,3) entry of A*B).
    
    To compute the dot product we see that there is only a non-zero contribution
    for index 4:
        
      A[1,0]*B[0,3] +A[1,1]*B[1,3] +A[1,2]*B[2,3] +A[1,3]*B[3,3] +A[1,4]*B[4,3]
    = 0*0           +0*0.7         +0*0           +0.3*0         + 0.8*0.1
    
    If we label the array of agreeing indices as agreements_a and agreements_b,
    the entry (A*B)[1,3] is therefore determined as
    
    dot(data_a[agreements_a], data_b[agreements_b])
    
    4. After having done this for all unique markers, we end up with a K times L
    array R, where K is the amount of unique markers in A and L is the amount of unique
    markers in B.
    
    To retrieve the COO representation of R, we first filter all non-zero indices.
    For this we can apply some floating point error tolerance mechanism.
    
    This will give as the index arrays I, J where R is non-zero.
    The rows array of the COO representation of A*B is now given as unique_marker_A[I].
    The colums array of the COO representation of A*B is now given as unique_marker_B[j].
    The data is R[I,J]

    """
    
    A_row, A_col, A_data = sort_indices_jitted(A_row, A_col, A_data, A_shape[1])
    B_col, B_row, B_data = sort_indices_jitted(B_col, B_row, B_data, B_shape[0])
    
    unique_marker_a = find_unique_markers(A_row)
    unique_marker_b = find_unique_markers(B_col)

    R = coo_mult_kernel(A_col, A_data, B_row, B_data, unique_marker_a, unique_marker_b)
    
    abs_R = np.abs(R)
    max_abs = np.max(R.ravel())
    
    I, J = np.nonzero(abs_R > (cutoff_ratio * max_abs))
    
    res_row = A_row[unique_marker_a[I]]
    res_col = B_col[unique_marker_b[J]]
    
    R_flat = R.ravel()
    res_data = R_flat[J + R.shape[1]*I]

    return res_row, res_col, res_data

@njit(parallel = True, cache = True)                
def coo_mult_kernel(A_col, A_data, B_row, B_data, unique_marker_a, unique_marker_b):
        
    res = np.zeros((len(unique_marker_a)-1, len(unique_marker_b)-1), dtype = A_data.dtype)
    
    for i in prange(len(unique_marker_a)-1):
        comparison_block_a = A_col[unique_marker_a[i]:unique_marker_a[i+1]]        
        for j in range(len(unique_marker_b)-1):
                
            if comparison_block_a[0] > B_row[unique_marker_b[j+1]-1]:
                continue
            
            if B_row[unique_marker_b[j]] > comparison_block_a[-1]:
                continue
            
            comparison_block_b = B_row[unique_marker_b[j]:unique_marker_b[j+1]]
            
            agreeing_ind_a, agreeing_ind_b = find_agreements(comparison_block_a, comparison_block_b)
            
            res[i, j] = np.dot(A_data[unique_marker_a[i] + agreeing_ind_a], B_data[unique_marker_b[j] + agreeing_ind_b])

    return res

@njit(cache = True)
def find_unique_markers(arr):
    
    unique_marker = [0]
    
    cur_value = arr[0]
    for i in range(len(arr)):
        if cur_value != arr[i]:
            unique_marker.append(i)
            cur_value = arr[i]
            
    unique_marker.append(len(arr))
            
    return np.array(unique_marker, dtype = np.int64)
    
@njit(cache = True)
def find_agreements(a, b):
    A = np.broadcast_to(b, (a.shape[0], b.shape[0]))
    B = np.broadcast_to(a, (b.shape[0], a.shape[0]))
    return np.nonzero(A==B.transpose())


def coo_sparse_matrix_mult(A, B):
    if A.shape[0] < B.shape[1]:
        new_row, new_col, new_data = coo_sparse_matrix_mult_inner(A.row, A.col, A.data, B.row, B.col, B.data, A.shape, B.shape)
        
    else:
        
        new_col, new_row, new_data = coo_sparse_matrix_mult_inner(B.col, B.row, B.data, A.col, A.row, A.data, B.shape[::-1], A.shape[::-1])
        
    if len(new_data) == 0:
        return coo_array(([], ([], [])), shape = (A.shape[0], B.shape[1]))

    if np.max(new_row) < 1<<31 and not new_row.dtype == np.int32:
        new_row = np.array(new_row, dtype = np.int32)
        
    if np.max(new_col) < 1<<31 and not new_col.dtype == np.int32:
        new_col = np.array(new_col, dtype = np.int32)
            
    return coo_array((new_data, (new_row, new_col)), shape = (A.shape[0], B.shape[1]))


def sparse_matrix_mult(A, B):
    
    if A.shape[0]*A.shape[1] < B.shape[0]*B.shape[1]:

        log_shape_0_a = np.log2(A.shape[0])
        log_shape_1_b = np.log2(B.shape[1])
        
        log_sparsity_a = -np.log2(A.nnz/(A.shape[0]*A.shape[1]))
        log_sparsity_b = -np.log2(B.nnz/(B.shape[0]*B.shape[1]))
        
    else:
        log_shape_0_a = np.log2(B.shape[1])
        log_shape_1_b = np.log2(A.shape[0])
        
        log_sparsity_a = -np.log2(B.nnz/(B.shape[0]*B.shape[1]))
        log_sparsity_b = -np.log2(A.nnz/(A.shape[0]*A.shape[1]))
    
    if get_prediction(log_shape_0_a, log_shape_1_b, log_sparsity_a, log_sparsity_b):
        return (A @ B).tocoo()
    else:
        return  coo_sparse_matrix_mult(A, B)
    

def get_prediction(log_shape_0_a, log_shape_1_b, log_sparsity_a, log_sparsity_b):
    
    data_tuple = (log_shape_0_a, log_shape_1_b, log_sparsity_a, log_sparsity_b)
    
    # Normalize and add bias to the input data
    x_mean = np.array([ 5.488, 28.741,  7.679, 26.106])
    x_std = np.array([ 2.26844793, 11.88553402,  3.73764083, 12.94182228])
    theta = np.array([-3.37310532,  1.05192584, -0.98338696, -0.6838338 , -2.10351854])
    normalized_data = (np.array(data_tuple) - x_mean) / x_std
    data_with_bias = np.hstack(([1], normalized_data))

    # Sigmoid function
    def sigmoid(z):
        return 1 / (1 + np.exp(-z))
    
    # Calculate the probability using the trained theta values
    probability = sigmoid(np.dot(data_with_bias, theta))

    # Make a binary prediction (True/False) based on the probability threshold (0.5)
    prediction = probability >= 0.5

    return prediction    


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import inspect

from qrisp.environments.quantum_environments import QuantumEnvironment
from qrisp.environments.gate_wrap_environment import GateWrapEnvironment
from qrisp.circuit import Operation, QuantumCircuit, Instruction
from qrisp.environments.iteration_environment import IterationEnvironment
from qrisp.core import merge

def custom_control(func):
    """
    The ``custom_control`` decorator allows to specify the controlled version of 
    the decorated function. If this function is called within a :ref:`ControlEnvironment`
    or a :ref:`ConditionEnvironment` the controlled version is executed instead.
    
    Specific controlled versions of quantum functions are very common in many
    scientific publications. This is because the general control procedure can
    signifcantly increase resource demands.
    
    In order to use the ``custom_control`` decorator, you need to add the ``ctrl``
    keyword to your function signature. If called within a controlled context,
    this keyword will receive the corresponding control qubit.
    
    For more details consult the examples section.
    

    Parameters
    ----------
    func : function
        A function of QuantumVariables, which has the ``ctrl`` keyword.

    Returns
    -------
    adaptive_control_function : function
        A function which will execute it's controlled version, if called
        within a :ref:`ControlEnvironment` or a :ref:`ConditionEnvironment`.
        
    Examples
    --------
    
    We create a swap function with custom control.
    
    ::
        
        from qrisp import mcx, cx, custom_control
        
        @custom_control
        def swap(a, b, ctrl = None):
            
            if ctrl is None:
                
                cx(a, b)
                cx(b, a)
                cx(a, b)
                
            else:
                
                cx(a, b)
                mcx([ctrl, b], a)
                cx(a, b)
                
    
    Test the non-controlled version:
        
    ::
        
        from qrisp import QuantumBool
        
        a = QuantumBool()
        b = QuantumBool()
        
        swap(a, b)
        
        print(a.qs)
        
    
    ::
        
        QuantumCircuit:
        --------------
                  ┌───┐     
        a.0: ──■──┤ X ├──■──
             ┌─┴─┐└─┬─┘┌─┴─┐
        b.0: ┤ X ├──■──┤ X ├
             └───┘     └───┘
        Live QuantumVariables:
        ---------------------
        QuantumBool a
        QuantumBool b        
        
    
    Test the controlled version:
        
    ::
        
        from qrisp import control
        
        a = QuantumBool()
        b = QuantumBool()
        ctrl_qbl = QuantumBool()
        
        with control(ctrl_qbl):
            
            swap(a,b)
            
        print(a.qs.transpile(1))
        
    ::
        
                         ┌───┐     
               a.0: ──■──┤ X ├──■──
                    ┌─┴─┐└─┬─┘┌─┴─┐
               b.0: ┤ X ├──■──┤ X ├
                    └───┘  │  └───┘
        ctrl_qbl.0: ───────■───────
    

    """
    
    def adaptive_control_function(*args, **kwargs):
        
        from qrisp.core import recursive_qs_search
        from qrisp import merge, ControlEnvironment, ConditionEnvironment, QuantumEnvironment, InversionEnvironment, ConjugationEnvironment
        
        qs_list = recursive_qs_search(args)
           
        if "ctrl" in kwargs:
            if kwargs["ctrl"] is not None:
                qs_list.append(kwargs["ctrl"].qs())
        
        merge(qs_list)
        
        if len(qs_list) == 0:
            return func(*args, **kwargs)
        
        qs = qs_list[0]
        
        # Search for a Control/Condition Environment and get the control qubit
        control_qb = None
        for env in qs.env_stack[::-1]:
            if type(env) == QuantumEnvironment:
                continue
            if isinstance(env, (ControlEnvironment, ConditionEnvironment)):
                control_qb = env.condition_truth_value
                break
            if not isinstance(env, (InversionEnvironment, ConjugationEnvironment, GateWrapEnvironment)):
                if isinstance(env, IterationEnvironment):
                    if env.precompile:
                        break
                    else:
                        continue
                break

        # If no control qubit was found, simply execute the function
        if control_qb is None:
            return func(*args, **kwargs)

        # Check whether the function supports the ctrl_method kwarg and adjust
        # the kwargs accordingly
        if "ctrl_method" in list(inspect.getfullargspec(func))[0] and isinstance(env, ControlEnvironment):
            kwargs.update({"ctrl_method" : env.ctrl_method})
        
        
        # In the case that a qubit was found, we use the CustomControlEnvironent (definded below)
        # This environments gatewraps the function and compiles it to a specific Operation subtype
        # called CustomControlledOperation.
        # The Condition/Control Environment compiler recognizes this Operation type
        # and processes it accordingly

        with CustomControlEnvironment(control_qb, func.__name__):
            if "ctrl" in kwargs:
                kwargs["ctrl"] = control_qb
                res = func(*args, **kwargs)
            else:
                res = func(*args, ctrl = control_qb, **kwargs)
        
        return res
        
    return adaptive_control_function


class CustomControlEnvironment(QuantumEnvironment):
    
    def __init__(self, control_qb, name):
        
        self.control_qb = control_qb
        self.manual_allocation_management = True
    
    def __enter__(self):
        
        QuantumEnvironment.__enter__(self)
        merge([self.env_qs, self.control_qb.qs()])
    
    def compile(self):
        
        
        original_data = list(self.env_qs.data)
        self.env_qs.data = []
        QuantumEnvironment.compile(self)
        
        temp = list(self.env_qs.data)
        self.env_qs.data = []
        
        for instr in temp:
            
            if instr.op.name in ["qb_alloc", "qb_dealloc"]:
                self.env_qs.append(instr)
            elif self.control_qb in instr.qubits:
                cusc_op = CustomControlOperation(instr.op, targeting_control = True)
                self.env_qs.append(cusc_op, instr.qubits, instr.clbits)
            else:
                cusc_op = CustomControlOperation(instr.op)
                self.env_qs.append(cusc_op, [self.control_qb] + instr.qubits, instr.clbits)
                
        self.env_qs.data = original_data + list(self.env_qs.data)
        

class CustomControlOperation(Operation):
    
    def __init__(self, init_op, targeting_control = False):
        
        self.targeting_control = targeting_control
        
        if not targeting_control:
            definition = QuantumCircuit(init_op.num_qubits + 1, init_op.num_clbits)
            definition.data.append(Instruction(init_op, definition.qubits[1:], definition.clbits))
            
            Operation.__init__(self, name = "cusc_" + init_op.name, num_qubits = init_op.num_qubits + 1, num_clbits = init_op.num_clbits, definition = definition)
            
            self.init_op = init_op
            
            self.permeability = {i+1 : init_op.permeability[i] for i in range(init_op.num_qubits)}
            self.permeability[0] = True
            
            self.is_qfree = init_op.is_qfree
        else:
            
            definition = QuantumCircuit(init_op.num_qubits, init_op.num_clbits)
            definition.append(init_op, definition.qubits, definition.clbits)
            
            Operation.__init__(self, name = "cusc_" + init_op.name, num_qubits = init_op.num_qubits, num_clbits = init_op.num_clbits, definition = definition)
            
            self.init_op = init_op
            
            self.permeability = dict(init_op.permeability)
            self.is_qfree = bool(init_op.is_qfree)
            

    def inverse(self):
        temp = self.init_op.inverse()
        return CustomControlOperation(temp, targeting_control = self.targeting_control)
    
    def copy(self):
        return CustomControlOperation(self.init_op.copy(), targeting_control = self.targeting_control)




"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import inpl_mult, h, QuantumFloat, qswitch, multi_measurement
def test_qswitch():

    # Some sample case functions
    def f0(x): x += 1
    def f1(x): inpl_mult(x, 3, treat_overflow = False)
    def f2(x): pass
    def f3(x): h(x[1])
    case_function_list = [f0, f1, f2, f3]
    
    # Create operand and case variable
    operand = QuantumFloat(4)
    operand[:] = 1
    case = QuantumFloat(2)
    h(case)
    
    # Execute switch_case function
    qswitch(operand, case, case_function_list)
    
    # Simulate
    assert multi_measurement([case, operand]) == {(0, 2): 0.25, (1, 3): 0.25, (2, 1): 0.25, (3, 1): 0.125, (3, 3): 0.125}
    # Yields {(0, 2): 0.25, (1, 3): 0.25, (2, 1): 0.25, (3, 1): 0.125, (3, 3): 0.125}
    
    # Create operand and case variable
    operand = QuantumFloat(4)
    operand[:] = 1
    case = QuantumFloat(2)
    h(case)
    
    # Execute switch_case function
    qswitch(operand, case, case_function_list, method = "parallel")
    
    # Simulate
    assert multi_measurement([case, operand]) == {(0, 2): 0.25, (1, 3): 0.25, (2, 1): 0.25, (3, 1): 0.125, (3, 3): 0.125}
    # Yields {(0, 2): 0.25, (1, 3): 0.25, (2, 1): 0.25, (3, 1): 0.125, (3, 3): 0.125}



"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.circuit.qubit import Qubit
from qrisp.operators.qubit.qubit_operator import QubitOperator
from qrisp.operators.qubit.bound_qubit_operator import BoundQubitOperator
from qrisp.operators.qubit.qubit_term import QubitTerm
from qrisp.operators.qubit.bound_qubit_term import BoundQubitTerm

def X(arg):
    if isinstance(arg,int):
        return QubitOperator({QubitTerm({arg:"X"}):1})
    elif isinstance(arg, Qubit):
        return BoundQubitOperator({BoundQubitTerm({arg:"X"}):1})
    else:
        raise Exception("Cannot initialize operator from type "+str(type(arg)))

def Y(arg):
    if isinstance(arg,int):
        return QubitOperator({QubitTerm({arg:"Y"}):1})
    elif isinstance(arg, Qubit):
        return BoundQubitOperator({BoundQubitTerm({arg:"Y"}):1})
    else:
        raise Exception("Cannot initialize operator from type "+str(type(arg)))

def Z(arg):
    if isinstance(arg,int):
        return QubitOperator({QubitTerm({arg:"Z"}):1})
    elif isinstance(arg, Qubit):
        return BoundQubitOperator({BoundQubitTerm({arg:"Z"}):1})
    else:
        raise Exception("Cannot initialize operator from type "+str(type(arg)))

def A(arg):
    return QubitOperator({QubitTerm({arg : "A"}): 1})

def C(arg):
    return QubitOperator({QubitTerm({arg : "C"}): 1})

def P0(arg):
    return QubitOperator({QubitTerm({arg : "P0"}): 1})

def P1(arg):
    return QubitOperator({QubitTerm({arg : "P1"}): 1})



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import *
import numpy as np
import sympy as sp
from operator import itemgetter

def QUBO_obj(bitstring, Q):
    x = np.array(list(bitstring), dtype=int)
    cost = x.T @ Q @ x
    return cost

def create_QUBO_cl_cost_function(Q):
    """
    Creates the classical cost function for a QUBO instance with matrix ``Q`` that we are attempting to solve.

    Parameters
    ----------
    Q : np.array
        QUBO matrix to solve.

    Returns
    -------
    cl_cost_function : function
        The classical cost function for the problem instance, which takes a dictionary of measurement results as input.

    """    
    def cl_cost_function(counts):
        
        def QUBO_obj(bitstring, Q):
            x = np.array(list(bitstring), dtype=int)
            cost = x.T @ Q @ x
            return cost
        
        energy = 0
        for meas, meas_count in counts.items():
            obj_for_meas = QUBO_obj(meas,Q)
            energy += obj_for_meas * meas_count
        return energy
    
    return cl_cost_function


# MAPPING QUBO TO GATES                       1 - Z_i                 1
#                               x_i  <-----> ---------:     x^T Q x = - SUM_{i,j}  Q_{ij} (1-Z_i)(1-Z_j) = SUM_{i,j} Q_ij*(Z_i*Z_j - Z_i - Z_j + 1)/4
#                                                2                    4  


def create_QUBO_cost_operator(Q):
    """
    Creates the cost operator for a QUBO instance with matrix ``Q``.
    In the QAOA overview section this is also called the phase separator $U_P$.

    Parameters
    ----------
    Q : np.array
        QUBO matrix to solve.

    Returns
    -------
    cost_operator : function
        A function receiving a :ref:`QuantumVariable` and a real parameter $\gamma$.
        This function performs the application of the cost operator.

    """
    
    def QUBO_cost_operator(qv, gamma):

        # Rescaling for enhancing the performance of the QAOA
        gamma = gamma/np.linalg.norm(Q)

        gphase(-gamma/4*(np.sum(Q)+np.trace(Q)),qv[0])

        for i in range(len(Q)):
            rz(-gamma/2*(sum(Q[i])+sum(Q[:,i])), qv[i])
            for j in range(len(Q)):
                if i != j and Q[i][j] != 0:
                    rzz(gamma/2*Q[i][j], qv[i], qv[j])

    return QUBO_cost_operator


def QUBO_problem(Q):
    """
    Creates a QAOA problem instance with appropriate phase separator, mixer, and
    classical cost function.

    Parameters
    ----------
    Q : np.array
        QUBO matrix to solve.

    Returns
    -------
    :ref:`QAOAProblem`
        A QAOA problem instance for a given QUBO matrix ``Q``.

    """    
    from qrisp.qaoa import QAOAProblem, RX_mixer
    
    return QAOAProblem(create_QUBO_cost_operator(Q), RX_mixer, create_QUBO_cl_cost_function(Q))


def solve_QUBO(Q, depth, shots=5000, max_iter = 50, backend = None, print_res = False):
    """
    Solves a Quadratic Unconstrained Binary Optimization (QUBO) problem using the Quantum Approximate Optimization Algorithm (QAOA). 

    This function creates the QAOA problem for a given QUBO. 
    It defines a quantum argument as a :ref:`QuantumArray` of ``len(Q)`` :ref:`QuantumVariables <QuantumVariable>` with size 1. 
    It then runs the QAOA with the given quantum argument, ``depth`` (number of layers), maximum ``iterations`` of the classical optimizer, and ``shots`` and ``backend`` as measurement keyword arguments. 
    The method performs classical post-processing on the solutions of the QAOA optimization: 
    it calculates the cost for each such solution, sorts the solutions by their cost in ascending order, and returns the optimal solution.

    .. warning::
     
        For small QUBO instance the number of ``shots`` typically exceeds the number of possible solutions.
        In this case, even QAOA with ``depth=0``, i.e., sampling from a uniform superposition, may yield the optimal solution as the classical post-processing amounts to brute force search!
        Performance of :meth:`solve_QUBO <qrisp.qaoa.problems.QUBO.solve_QUBO>` for small instance may not be indicative of performance for large instances. 

    Parameters
    ----------
    Q : np.array
        QUBO matrix to solve.
    depth : int
        The depth (amount of layers) of the QAOA circuit.
    shots : int
        The number of shots. The default is 5000.
    max_iter : int, optional
        The maximal amount of iterations of the ``COBYLA`` optimizer in the QAOA algorithm.
        The default is 50.
    backend : :ref:`BackendClient`, optional
        The backend to be used for the quantum simulation. 
        By default, the Qrisp simulator is used.
    print_res : Boolean, optional
        If set to ``True``, the 5 best solutions are printed. The default is ``False``.

    Returns
    -------
    optimal_solution : tuple
        A tuple where the first element is the cost value and the second element is the optimal bitstring. 
        If ``print_res`` is set to ``True``, the function prints the 5 best solutions with their respective costs.

    Examples
    --------

    ::

        from qrisp.qaoa import solve_QUBO
        import numpy as np

        Q = np.array(
            [
                [-17,  10,  10,  10,   0,  20],
                [ 10, -18,  10,  10,  10,  20],
                [ 10,  10, -29,  10,  20,  20],
                [ 10,  10,  10, -19,  10,  10],
                [ 0,   10,  20,  10, -17,  10],
                [ 20,  20,  20,  10,  10, -28],
            ]
        )

        solve_QUBO(Q, depth = 1, print_res=True)

    """

    # Define quantum argument as a QuantumArray of len(G) QuantumVariables with size 1 or as a QuantumVariable with size len(G)
    qarg = QuantumArray(qtype = QuantumVariable(1), shape = len(Q))

    QUBO_instance = QUBO_problem(Q)

    if backend is None:
        from qrisp.default_backend import def_backend
        backend = def_backend
    else:
        backend = backend

    # Run QAOA with given quantum arguments, depth, measurement keyword arguments and maximum iterations for optimization
    res = QUBO_instance.run(qarg, depth, mes_kwargs={"backend" : backend, "shots" : shots}, max_iter = max_iter, init_type='tqa') # runs the simulation

    # Calculate the cost for each solution
    costs_and_solutions = [(QUBO_obj(bitstring, Q), bitstring) for bitstring in res.keys()]

    # Sort the solutions by their cost in ascending order
    sorted_costs_and_solutions = sorted(costs_and_solutions, key=itemgetter(0))

    optimal_solution = sorted_costs_and_solutions[0]

    if print_res is True:
        # Get the best solutions and print them
        for i in range(5):
            print(f"Solution {i+1}: {sorted_costs_and_solutions[i][1]} with cost: {sorted_costs_and_solutions[i][0]} and probability: {res[sorted_costs_and_solutions[i][1]]}")
    
    return optimal_solution




"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

from qrisp.operators import Hamiltonian
from qrisp.operators.fermionic.fermionic_term import FermionicTerm
from qrisp.operators.hamiltonian_tools import group_up_terms
from qrisp import merge, IterationEnvironment, conjugate
from qrisp.operators.qubit import QubitOperator

import sympy as sp

threshold = 1e-9

#
# FermionicOperator
#

class FermionicOperator(Hamiltonian):
    r"""
    This class provides an efficient implementation of ladder_term operators, i.e.,
    operators of the form

    .. math::
        
        O=\sum\limits_{j}\alpha_jO_j 
            
    where each term $O_j$ is a product of fermionic raising $a_i^{\dagger}$ and lowering $a_i$ operators acting on the $i$ th fermionic mode.

    The ladder operators satisfy the commutation relations

    .. math::

        \{a_i,a_j^{\dagger}\} &= a_ia_j^{\dagger}+a_j^{\dagger}a_i = \delta_{ij}\\
        \{a_i^{\dagger},a_j^{\dagger}\} &= \{a_i,a_j\} = 0

    Examples
    --------

    A ladder term operator can be specified conveniently in terms of ``a`` (lowering, i.e., annihilation), ``c`` (raising, i.e., creation) operators:

    ::
        
        from qrisp.operators.fermionic import a, c

        O = a(2)*c(1)+a(3)*c(2)
        O

    Yields $a_2c_1+a_3c_2$.

    """

    def __init__(self, terms_dict={}):
        
        self.terms_dict = dict(terms_dict)

    def reduce(self, assume_hermitian = False):
        """
        Applies the fermionic anticommutation laws to bring the operator into
        a standard form. This can reduce the amount of terms because several
        terms might be the permuted version of each other and therefore their
        coefficients add up.
        
        This function can reduce the amount of terms even further if the user
        can guarantee that the operator will be hermitized. In this case more
        identifications can be made.

        Parameters
        ----------
        assume_hermitian : bool, optional
            If set to True the function will assume that the result will be 
            hermitized. The default is False.

        Returns
        -------
        FermionicOperator
            The reduced FermionicOperator.

        Examples
        --------
        
        We create a FermionicOperator with redundant term definitions:
            
        ::
            
            from qrisp.operators import *
            
            O = a(0)*a(1) - a(1)*a(0)
            print(O.reduce())
            # Yields: 2*a0*a1

        To demonstrate the ``assume_hermitian`` feature, we create a FermionicOperator
        that has redundant terms, if hermitized.
        
            
        >>> O = a(0)*a(1) + c(1)*c(0)
        >>> reduced_O = O.reduce(assume_hermitian = True)
        >>> print(reduced_O)
        2*a0*a1
        
        Hermitizing gives the original operator.
        
        >>> print(reduced_O.hermitize())
        1.0*a0*a1 + 1.0*c1*c0
            
        """
        
        # This function performs some non trivial logic.
        
        # The problem here is that each fermionic term can
        # be reshaped into several different forms and still express
        # the same operator. 
        # For instance, the operator can be arbitrarily reordered
        # if the permutation sign is take care of and no creators/annihilators
        # with the same index are swapped.
        # Furthermore the Hamiltonian must be Hermitian, so each
        # term must be equivalent to is Hermitian conjugate.
        
        # This function implements a storage system, that combines
        # the coefficients of differing terms representing the same operator
        # into a single term.
        
        # This dictionary will contain the new terms, where redundancies
        # are taken care of.
        new_terms_dict = {}
        
        for term, coeff in self.terms_dict.items():
            
            # We only store the sorted version of each term.
            # Sorting here means permuting the creators/annihilators
            # while considering the sign of the permutation applied by the sort.
            # The sort is performed in a stable manner, so terms like a(0)*c(0)
            # don't get permuted (this would be a non-trivial anti-commutator).
            sorted_term, flip_sign = term.sort()
            if sorted_term not in new_terms_dict and assume_hermitian:
                # If the sorted term is not in the terms dict, the sorted version
                # of the daggering might be.
                daggered_sorted_term, daggered_flip_sign = term.dagger().sort()
                if daggered_sorted_term in new_terms_dict:
                    sorted_term = daggered_sorted_term
                    flip_sign = daggered_flip_sign
                                
            # Compute the new coefficient.
            new_terms_dict[sorted_term] = flip_sign*coeff + new_terms_dict.get(sorted_term, 0)
            
        return FermionicOperator(new_terms_dict)

    def len(self):
        return len(self.terms_dict)
    
    #
    # Printing
    #

    def _repr_latex_(self):
        # Convert the sympy expression to LaTeX and return it
        expr = self.to_expr()
        return f"${sp.latex(expr)}$"
    
    def __str__(self):
        # Convert the sympy expression to a string and return it
        expr = self.to_expr()
        return str(expr)
    
    def to_expr(self):
        """
        Returns a SymPy expression representing the operator.

        Returns
        -------
        expr : sympy.expr
            A SymPy expression representing the operator.

        """
        
        expr = 0  
        for ladder_term,coeff in self.terms_dict.items():
            expr += coeff*ladder_term.to_expr()
        return expr

    #
    # Arithmetic
    #
    
    def dagger(self):
        r"""
        Returns the daggered/adjoint version of self.

        Returns
        -------
        FermionicOperator
            The Operator $O^\dagger$.

        Examples
        --------
        
        We create a FermionicOperator and dagger it:
            
        ::
            
            from qrisp.operators import *
            
            O = a(0)*c(1)*a(2) + a(3)
            print(O.dagger())
            # Yields: c2*a1*c0 + c3
        """
        terms_dict = {}
        for term, coeff in self.terms_dict.items():
            terms_dict[term.dagger()] = np.conj(coeff)
        return FermionicOperator(terms_dict)    
    
    def hermitize(self):
        r"""
        Returns the hermitized version of self.

        Returns
        -------
        FermionicOperator
            The Operator $(O + O^\dagger)/2$.

        Examples
        --------
        
        We create a FermionicOperator and hermitize it:
            
        ::
            
            from qrisp.operators import *
            
            O = a(0)*c(1)*a(2) + a(3)
            print(O.hermitize())
            # Yields: 0.5*a0*c1*a2 + 0.5*a3 + 0.5*c2*a1*c0 + 0.5*c3
            
        """
        return 0.5*(self + self.dagger())
    
    def __eq__(self, other):
        reduced_self = self.reduce()
        reduced_other = other.reduce()
        
        if len(reduced_self.terms_dict) != len(reduced_other.terms_dict):
            return False
        
        for term, coeff in reduced_self.terms_dict.items():
            if not term in other.terms_dict:
                daggered_sorted_term, flip_sign = term.dagger().sort()
                if daggered_sorted_term not in reduced_other.terms_dict:
                    return False
                elif reduced_self.terms_dict[term] != flip_sign*reduced_other.terms_dict[daggered_sorted_term]:
                    return False
                continue
                    
            if reduced_self.terms_dict[term] != reduced_other.terms_dict[term]:
                return False
        
        return True
    
    def __neg__(self):
        return -1*self
            
    #def __pow__(self, e):
    #    if self.len()==1:
    #        if isinstance(e, int) and e>=0:
    #            if e%2==0:
    #                return FermionicOperator({FermionicTerm():1})
    #            else:
    #                return self
    #        else:
    #            raise TypeError("Unsupported operand type(s) for ** or pow(): "+str(type(self))+" and "+str(type(e)))
    #    else:
    #        raise TypeError("Unsupported operand type(s) for ** or pow(): "+str(type(self))+" and "+str(type(e)))

    def __add__(self,other):
        """
        Returns the sum of the operator self and other.

        Parameters
        ----------
        other : int, float, complex or FermionicOperator
            A scalar or a FermionicOperator to add to the operator self.

        Returns
        -------
        result : FermionicOperator
            The sum of the operator self and other.

        """

        if isinstance(other,(int,float,complex)):
            other = FermionicOperator({FermionicTerm():other})
        if not isinstance(other,FermionicOperator):
            raise TypeError("Cannot add FermionicOperator and "+str(type(other)))

        res_terms_dict = {}

        for ladder_term,coeff in self.terms_dict.items():
            res_terms_dict[ladder_term] = res_terms_dict.get(ladder_term,0)+coeff
            if abs(res_terms_dict[ladder_term])<threshold:
                del res_terms_dict[ladder_term]
    
        for ladder_term,coeff in other.terms_dict.items():
            res_terms_dict[ladder_term] = res_terms_dict.get(ladder_term,0)+coeff
            if abs(res_terms_dict[ladder_term])<threshold:
                del res_terms_dict[ladder_term]
        
        result = FermionicOperator(res_terms_dict)
        return result
    
    def __sub__(self,other):
        """
        Returns the difference of the operator self and other.

        Parameters
        ----------
        other : int, float, complex or FermionicOperator
            A scalar or a FermionicOperator to substract from the operator self.

        Returns
        -------
        result : FermionicOperator
            The difference of the operator self and other.

        """

        if isinstance(other,(int,float,complex)):
            other = FermionicOperator({FermionicTerm():other})
        if not isinstance(other,FermionicOperator):
            raise TypeError("Cannot substract FermionicOperator and "+str(type(other)))

        res_terms_dict = {}

        for ladder_term,coeff in self.terms_dict.items():
            res_terms_dict[ladder_term] = res_terms_dict.get(ladder_term,0)+coeff
            if abs(res_terms_dict[ladder_term])<threshold:
                del res_terms_dict[ladder_term]
    
        for ladder_term,coeff in other.terms_dict.items():
            res_terms_dict[ladder_term] = res_terms_dict.get(ladder_term,0)-coeff
            if abs(res_terms_dict[ladder_term])<threshold:
                del res_terms_dict[ladder_term]
        
        result = FermionicOperator(res_terms_dict)
        return result
    
    def __rsub__(self,other):
        """
        Returns the difference of the operator other and self.

        Parameters
        ----------
        other : int, float, complex or FermionicOperator
            A scalar or a FermionicOperator to substract from the operator self from.

        Returns
        -------
        result : FermionicOperator
            The difference of the operator other and self.

        """

        if isinstance(other,(int,float,complex)):
            other = FermionicOperator({FermionicTerm():other})
        if not isinstance(other,FermionicOperator):
            raise TypeError("Cannot substract FermionicOperator and "+str(type(other)))

        res_terms_dict = {}

        for ladder_term,coeff in self.terms_dict.items():
            res_terms_dict[ladder_term] = res_terms_dict.get(ladder_term,0)-coeff
            if abs(res_terms_dict[ladder_term])<threshold:
                del res_terms_dict[ladder_term]
    
        for ladder_term,coeff in other.terms_dict.items():
            res_terms_dict[ladder_term] = res_terms_dict.get(ladder_term,0)+coeff
            if abs(res_terms_dict[ladder_term])<threshold:
                del res_terms_dict[ladder_term]
        
        result = FermionicOperator(res_terms_dict)
        return result

    def __mul__(self,other):
        """
        Returns the product of the operator self and other.

        Parameters
        ----------
        other : int, float, complex or FermionicOperator
            A scalar or a FermionicOperator to multiply with the operator self.

        Returns
        -------
        result : FermionicOperator
            The product of the operator self and other.

        """

        if isinstance(other,(int,float,complex)):
            other = FermionicOperator({FermionicTerm():other})
        if not isinstance(other,FermionicOperator):
            raise TypeError("Cannot multipliy FermionicOperator and "+str(type(other)))

        res_terms_dict = {}

        for ladder_term1, coeff1 in self.terms_dict.items():
            for ladder_term2, coeff2 in other.terms_dict.items():
                curr_ladder_term = ladder_term1*ladder_term2
                res_terms_dict[curr_ladder_term] = res_terms_dict.get(curr_ladder_term,0) + coeff1*coeff2

        result = FermionicOperator(res_terms_dict)
        return result

    __radd__ = __add__
    __rmul__ = __mul__

    #
    # Inplace arithmetic
    #

    def __iadd__(self,other):
        """
        Adds other to the operator self.

        Parameters
        ----------
        other : int, float, complex or FermionicOperator
            A scalar or a FermionicOperator to add to the operator self.

        """

        if isinstance(other,(int,float,complex)):
            self.terms_dict[FermionicTerm()] = self.terms_dict.get(FermionicTerm(),0)+other
            return self
        if not isinstance(other,FermionicOperator):
            raise TypeError("Cannot add FermionicOperator and "+str(type(other)))

        for ladder_term,coeff in other.terms_dict.items():
            self.terms_dict[ladder_term] = self.terms_dict.get(ladder_term,0)+coeff
            if abs(self.terms_dict[ladder_term])<threshold:
                del self.terms_dict[ladder_term]
        self.terms_dict = FermionicOperator(self.terms_dict).terms_dict
        return self         

    def __isub__(self,other):
        """
        Substracts other from the operator self.

        Parameters
        ----------
        other : int, float, complex or FermionicOperator
            A scalar or a FermionicOperator to substract from the operator self.

        """

        if isinstance(other,(int,float,complex)):
            self.terms_dict[FermionicTerm()] = self.terms_dict.get(FermionicTerm(),0)-other
            return self
        if not isinstance(other,FermionicOperator):
            raise TypeError("Cannot add FermionicOperator and "+str(type(other)))

        for ladder_term,coeff in other.terms_dict.items():
            self.terms_dict[ladder_term] = self.terms_dict.get(ladder_term,0)-coeff
            if abs(self.terms_dict[ladder_term])<threshold:
                del self.terms_dict[ladder_term]  
        return self
    
    def __imul__(self,other):
        """
        Multiplys other to the operator self.

        Parameters
        ----------
        other : int, float, complex or FermionicOperator
            A scalar or a FermionicOperator to multiply with the operator self.

        """

        if isinstance(other,(int,float,complex)):
            other = FermionicOperator({FermionicTerm():other})
        if not isinstance(other,FermionicOperator):
            raise TypeError("Cannot multipliy FermionicOperator and "+str(type(other)))

        res_terms_dict = {}

        for ladder_term1, coeff1 in self.terms_dict.items():
            for ladder_term2, coeff2 in other.terms_dict.items():
                curr_ladder_term = ladder_term1*ladder_term2
                res_terms_dict[curr_ladder_term] = res_terms_dict.get(curr_ladder_term,0) + coeff1*coeff2

        self.terms_dict = res_terms_dict    

    #
    # Miscellaneous
    #

    def apply_threshold(self,threshold):
        """
        Removes all ladder_term terms with coefficient absolute value below the specified threshold.

        Parameters
        ----------
        threshold : float
            The threshold for the coefficients of the ladder_term terms.

        """

        delete_list = []
        for ladder_term,coeff in self.terms_dict.items():
            if abs(coeff)<threshold:
                delete_list.append(ladder_term)
        for ladder_term in delete_list:
            del self.terms_dict[ladder_term]

    def to_sparse_matrix(self, mapping_type = "jordan_wigner"):
        """
        Returns a matrix representing the operator.
    
        Returns
        -------
        M : scipy.sparse.csr_matrix
            A sparse matrix representing the operator.
        mapping_type : string, optional
            How to embedd the fermionic terms into a QubitOperator. Currently 
            only ``jordan_wigner`` is supported.

        """
        return self.to_qubit_operator(mapping_type=mapping_type).to_sparse_matrix()

    def ground_state_energy(self):
        """
        Calculates the ground state energy (i.e., the minimum eigenvalue) of the operator classically.
    
        Returns
        -------
        E : float
            The ground state energy. 

        """
        return self.to_qubit_operator().ground_state_energy()
    
    @classmethod
    def from_pyscf(self, pyscf_molecular_data):
        """
        .. _pscf_loading:
            
        Loads the data of a `PySCF molecule <https://pyscf.org/user/gto.html>`_ 
        into a FermionicOperator.

        Parameters
        ----------
        pyscf_molecular_data : pyscf.gto.mole.Mole
            The molecule to load.

        Returns
        -------
        molecule_hamiltonian : FermionicOperator
            The molecule as an operator.
            
        Examples
        --------
        
        We load the Hydrogen molecule and perform a hamiltonian simulation:

        >>> from pyscf import gto
        >>> mol = gto.M(atom = '''H 0 0 0; H 0 0 0.74''', basis = 'sto-3g')
        >>> H = FermionicOperator.from_pyscf(mol)
        >>> print(H)
        -0.181210462015197*a0*a1*c2*c3 + 0.181210462015197*a0*c1*c2*a3 
        - 1.25330978664598*c0*a0 + 0.674755926814448*c0*a0*c1*a1 
        + 0.482500939335616*c0*a0*c2*a2 + 0.663711401350814*c0*a0*c3*a3 
        + 0.181210462015197*c0*a1*a2*c3 - 0.181210462015197*c0*c1*a2*a3
        - 1.25330978664598*c1*a1 + 0.663711401350814*c1*a1*c2*a2 
        + 0.482500939335616*c1*a1*c3*a3 - 0.475068848772178*c2*a2 
        + 0.697651504490463*c2*a2*c3*a3 - 0.475068848772178*c3*a3
        
        Create a :ref:`QuantumVariable` and initialize two electrons in the upper
        orbitals. 

        >>> from qrisp import QuantumVariable
        >>> electron_state = QuantumVariable(4)
        >>> electron_state[:] = "0011"
        
        Simulate for $t = 100$ `Angstrom seconds <https://en.wikipedia.org/wiki/Angstrom>`_.
        
        >>> U = H.trotterization()
        >>> U(electron_state, t = 100, steps = 20)
        >>> print(electron_state)
        {'0011': 0.75331, '1100': 0.24669}
        
        We see that the electrons decayed to one of the lower levels. How cool is that?!
        
        """
        from qrisp.algorithms.vqe.problems.electronic_structure import create_electronic_hamiltonian
        return create_electronic_hamiltonian(pyscf_molecular_data)

    #
    # Transformations
    #

    def to_qubit_operator(self, mapping_type='jordan_wigner'):
        """
        Transforms the FermionicOperator to a :ref:`QubitOperator`.

        Parameters
        ----------
        mapping : str, optional
            The mapping to transform the Hamiltonian. Available is ``jordan_wigner``.
            The default is ``jordan_wigner``.

        Returns
        -------
        O : :ref:`QubitOperator`
            The resulting QubitOperator.
            
        Examples
        --------
        
        We map a singular fermionic ladder operator to a QubitOperator to see
        the Jordan-Wigner embedding.
            
        >>> from qrisp.operators import a
        >>> O = a(4)
        >>> print(O.to_qubit_operator())
        Z_0*Z_1*Z_2*Z_3*A_4
        
        """
        
        if mapping_type=="jordan_wigner":
            res = QubitOperator({})
            for term, coeff in self.terms_dict.items():
                res += coeff*term.to_qubit_term(mapping_type="jordan_wigner")
            return res
        else:
            raise Exception(f"Don't know fermionic mapping {mapping_type}.")
    
    #
    # Measurement
    #

    def get_measurement(
        self,
        qarg,
        mapping_type="jordan_wigner",
        **measurement_kwargs
    ):
        r"""
        This method returns the expected value of a Hamiltonian for the state 
        of a quantum argument. Note that this method measures the **hermitized**
        version of the operator:
            
        .. math::
            
            H = (O + O^\dagger)/2

        Parameters
        ----------
        qarg : QuantumVariable or list[Qubit]
            The quantum argument to evaluate the Hamiltonian on.
        mapping_type : str
            The strategy on how to map the FermionicOperator to a QubitOperator. Default is ``jordan_wigner``
        measurement_kwargs : dict
            The keyword arguments of :meth:`QubitOperator.get_measurement`.

        Raises
        ------
        Exception
            If the containing QuantumSession is in a quantum environment, it is not
            possible to execute measurements.

        Returns
        -------
        float
            The expected value of the Hamiltonian.
            
        Examples
        --------

        We create a FermionicOperator and perform a measurement.
        
        >>> from qrisp.operators import *
        >>> from qrisp import QuantumVariable
        >>> qv = QuantumVariable(4)
        >>> O = a(0)*a(1) + a(2)*c(1) + c(2)*a(3)
        >>> print(O.get_measurement(qv))
        -0.007968127490039834

        """
        qubit_operator = self.to_qubit_operator(mapping_type)
        return qubit_operator.get_measurement(qarg, **measurement_kwargs)
        
    #
    # Trotterization
    #
    

    def trotterization(self, t = 1, steps = 1, iter = 1):
        r"""
        Returns a function for performing Hamiltonian simulation, i.e., approximately implementing the unitary operator $e^{itH}$ via Trotterization.
        Note that this method will always simulate the **hermitized** operator, i.e.
        
        .. math::
            
            H = (O + O^\dagger)/2

        Returns
        -------
        U : function 
            A Python function that implements the first order Suzuki-Trotter formula.
            Given a Hamiltonian $H=H_1+\dotsb +H_m$ the unitary evolution $e^{itH}$ is 
            approximated by 
            
            .. math::

                e^{itH}\approx U_1(t,N)=\left(e^{iH_1t/N}\dotsb e^{iH_mt/N}\right)^N

            This function recieves the following arguments:

            * qarg : QuantumVariable or QuantumArray
                The quantum argument.
            * t : float, optional
                The evolution time $t$. The default is 1.
            * steps : int, optional
                The number of Trotter steps $N$. The default is 1.
            * iter : int, optional 
                The number of iterations the unitary $U_1(t,N)$ is applied. The default is 1.
                
        Examples
        --------
        
        We simulate a simple FermionicOperator.
        
        >>> from sympy import Symbol
        >>> from qrisp.operators import a,c
        >>> from qrisp import QuantumVariable
        >>> O = a(0)*a(1) + a(2)
        >>> U = O.trotterization()
        >>> qv = QuantumVariable(3)
        >>> t = Symbol("t")
        >>> U(qv, t = t)
        >>> print(qv.qs)
        QuantumCircuit:
        ---------------
                    ┌───┐    ┌───┐          ┌────────────┐     ┌───┐┌───┐      
        qv.0: ────■─┤ X ├────┤ X ├───────■──┤ Rz(-0.5*t) ├──■──┤ X ├┤ X ├─■────
                  │ └─┬─┘    ├───┤     ┌─┴─┐├────────────┤┌─┴─┐├───┤└─┬─┘ │    
        qv.1: ─■──┼───■──────┤ H ├─────┤ X ├┤ Rz(-0.5*t) ├┤ X ├┤ H ├──■───┼──■─
               │  │ ┌───┐┌───┴───┴────┐├───┤└────────────┘└───┘└───┘      │  │ 
        qv.2: ─■──■─┤ H ├┤ Rz(-1.0*t) ├┤ H ├──────────────────────────────■──■─
                    └───┘└────────────┘└───┘                                   
        Live QuantumVariables:
        ----------------------
        QuantumVariable qv
        
        Execute a simulation:
            
        >>> print(qv.get_measurement(subs_dic = {t : 0.5}))
        {'000': 0.9242, '001': 0.06026, '110': 0.01459, '111': 0.00095}
            
        """
        
        reduced_H = self.reduce(assume_hermitian=True)
        
        groups = reduced_H.group_up(denominator = lambda a,b : not a.intersect(b))
        
        
        def trotter_step(qarg, t, steps):
            
            for group in groups:
                
                permutation = []
                terms = group.terms_dict.keys()
                for term in terms:
                    for ladder in term.ladder_list:
                        if ladder[0] not in permutation:
                            permutation.append(ladder[0])
                
                for k in range(len(qarg)):
                    if k not in permutation:
                        permutation.append(k)
                
                permutation = permutation[::-1]
                
                with conjugate(apply_fermionic_swap)(qarg, permutation) as new_qarg:
                    
                    for ferm_term in terms:
                        coeff = reduced_H.terms_dict[ferm_term]
                        pauli_hamiltonian = ferm_term.fermionic_swap(permutation).to_qubit_term()
                        pauli_term = list(pauli_hamiltonian.terms_dict.keys())[0]
                        pauli_term.simulate(coeff*t/steps*pauli_hamiltonian.terms_dict[pauli_term], new_qarg)
                

        def U(qarg, t=1, steps=1, iter=1):
            merge([qarg])
            with IterationEnvironment(qarg.qs, iter*steps):
                trotter_step(qarg, t, steps)

        return U
    
    def group_up(self, denominator):
        
        term_groups = group_up_terms(self, denominator)
        groups = []
        for term_group in term_groups:
            O = FermionicOperator({term : self.terms_dict[term] for term in term_group})
            groups.append(O)
            
        return groups
    
                    
def apply_fermionic_swap(qv, permutation):
    from qrisp import cz
    qb_list = list(qv)
    swaps = get_swaps_for_permutation(permutation)
    for swap in swaps[::-1]:
        cz(qb_list[swap[0]], qb_list[swap[1]])
        qb_list[swap[0]], qb_list[swap[1]] = qb_list[swap[1]], qb_list[swap[0]]
        
    return qb_list
        
def get_swaps_for_permutation(permutation):
    swaps = []
    permutation = list(permutation)
    for i in range(len(permutation)):
        j = permutation.index(i)
        while j != i:
            permutation[j], permutation[j-1] = permutation[j-1], permutation[j]
            swaps.append((j, j-1))
            j -= 1
    return swaps


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import qiskit

from qrisp.interface import VirtualBackend

def IQMBackend(api_token, device_instance):
    """
    This function instantiates an IQMBackend based on VirtualBackend
    using Qiskit and Qiskit-on-IQM.


    Parameters
    ----------
    api_token : str
        An API token retrieved from the IQM Resonance website.
    device_instance : str
        The device instance of the IQM backend such as "garnet".
        For an up-to-date list, see the IQM Resonance website.

    Examples
    --------

    We evaluate a QuantumFloat multiplication on the 20-qubit IQM Garnet.

    >>> from qrisp.interface import IQMBackend
    >>> qrisp_garnet = IQMBackend(api_token = "YOUR_IQM_RESONANCE_TOKEN", device_instance = "garnet")
    >>> from qrisp import QuantumFloat
    >>> a = QuantumFloat(2)
    >>> a[:] = 2
    >>> b = a*a
    >>> b.get_measurement(backend = qrisp_garnet, shots=1000)
    {4: 0.548,
     5: 0.082,
     0: 0.063,
     6: 0.042,
     8: 0.031,
     2: 0.029,
     12: 0.014,
     10: 0.03,
     1: 0.027,
     7: 0.025,
     15: 0.023,
     9: 0.021,
     14: 0.021,
     13: 0.018,
     11: 0.014,
     3: 0.012}

    """

    if not isinstance(api_token, str):
        raise TypeError(
            "api_token must be a string. You can create an API token on the IQM Resonance website."
        )

    if not isinstance(device_instance, str):
        raise TypeError(
            "Please provide a device_instance as a string. You can retrieve a list of available devices id on the IQM Resonance website."
        )

    try:
        from iqm.qiskit_iqm import IQMProvider, transpile_to_IQM
    except ImportError:
        raise ImportError(
            "Please install qiskit-iqm to use the IQMBackend. You can do this by running `pip install qrisp[iqm]`."
        )


    def run_func_iqm(qasm_str, shots=None, token=""):
        if shots is None:
            shots = 1000
        server_url = "https://cocos.resonance.meetiqm.com/" + device_instance

        backend = IQMProvider(server_url, token=api_token).get_backend()
        qc = qiskit.QuantumCircuit.from_qasm_str(qasm_str)
        qc_transpiled = transpile_to_IQM(qc, backend)

        job = backend.run(qc_transpiled, shots=shots)
        import re

        counts = job.result().get_counts()
        new_counts = {}
        for key in counts.keys():
            counts_string = re.sub(r"\W", "", key)
            new_counts[counts_string] = counts[key]

        return new_counts

    return VirtualBackend(run_func_iqm)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# This file implements the Gidney-Adder https://arxiv.org/abs/1709.06648
# but in a semi-classical manner. That is, the adder accepts one classical
# and one quantum input.

# For the quantum-quantum version please check qrisp.alg_primitives.arithmetic.adders.gidney_adder


from qrisp.qtypes import QuantumBool
from qrisp.core import QuantumVariable, merge
from qrisp.misc.utility import bin_rep
from qrisp.alg_primitives import x, cx, mcx
from qrisp.circuit import fast_append
from qrisp.environments import invert


# This function inverts a bit represented by a character
def c_inv(c):
    if c == "0":
        return "1"
    elif c == "1":
        return "0"
    else:
        raise Exception

# This function perform semi-classical Gidney-Addition

def cq_gidney_adder(a, b, c_in = None, c_out = None, ctrl = None):
    
    # Call merge to make the QuantumSession enter the inversion environment
    # (doesn't happen automatically with fast_append = 3)
    
    qs_list = [b[0].qs()]
    
    for qb in [c_in, c_out, ctrl]:
        if qb is not None:
            qs_list.append(qb.qs())
    
    merge(qs_list)
    
    with fast_append(3):
    
        if isinstance(a, int):
            a = bin_rep(a%2**len(b), len(b))[::-1]
        elif not isinstance(a, str):
            raise Exception(f"Tried to call semi-classical carry calculator with invalid type {type(a)}")
        
        if len(a) != len(b):
            raise Exception("Tried to call Gidney adder with inputs of unequal length")
    
        if c_out is not None:
            
            # Convert to qubit if neccessary
            if isinstance(c_out, QuantumBool):
                c_out = c_out[0]
            
            b = list(b) + [c_out]
            a = a + "0"
            
        if c_in is not None:
            
            # Convert to qubit if neccessary
            if isinstance(c_in, QuantumBool):
                c_in = c_in[0]
            
            b = [c_in] + list(b)
            a = "1" + a
            
        #Treat the case that b is only a single qubit
        if len(b) == 1:
            if a[0] == "1":
                if ctrl is None:
                    x(b[0])
                else:
                    cx(ctrl, b[0])
            return
        
        gidney_anc = QuantumVariable(len(b) - 1, name = "gidney_anc*", qs = b[0].qs())
        
        # This loop now perform the iterations that are required to build up the circuit described
        # in the paper
        for i in range(len(b)-1):
            
            # The relevant observation to turn Gidney's adder into a semi-classical adder
            # is that the "a_i" qubit (in his paper these qubits are called "i") 
            # is only involved in two interactions:
            #   1. The CNOT gate from the ancilla qubit of the previous iteration
            #   2. The the quasi-toffoli gate into the ancilla of the current interaction
            # If we consider "a_i" as a classically known value, we can recover the quantum
            # value of the first step by applying an x gate onto the ancilla of the previous 
            # iteration if a_i is True.
            
            # To "simulate" the quasi-toffoli we then simply use this ancilla as a control
            # value instead.
            
            # Combining these steps results in a quasi-toffoli with a modified control-state
            # If a is True, the quasi-toffoli receives the flipped value of the ancilla
            if i != 0:
                
                if ctrl is not None:
                    if "1" == a[i]:
                        cx(ctrl, gidney_anc[i-1])
                    mcx([gidney_anc[i-1], b[i]], gidney_anc[i], method = "gidney", ctrl_state = "11")
                    if "1" == a[i]:
                        cx(ctrl, gidney_anc[i-1])
                else:
                    mcx([gidney_anc[i-1], b[i]], gidney_anc[i], method = "gidney", ctrl_state = c_inv(a[i]) + "1")
                    
                cx(gidney_anc[i-1], gidney_anc[i])
                
            else:
                # For the first iteration, we can simply execute a CNOT gate if a_0 is True
                if a[i] == "1":
                    if ctrl is None:
                        cx(b[i], gidney_anc[i])
                    else:
                        mcx([ctrl, b[i]], gidney_anc[i], method = "gidney")
            
            
            # To realize the controlled version of this adder, note that the V-shaped
            # part of the circuit in the paper is consisting of two branches that are
            # almost inverse to each other.
            
            # The one thing that stops the from being truly inverse is the CNOT gate
            # from the ancilla into b (in his paper "t").
            
            # That is: If we only control this gate (and the "tip"), the two branches will be their
            # perfect inverse and the whole computation will result in the identity
            # if the control qubit is in the |0> state.
            if i != len(b) -2:
                cx(gidney_anc[i], b[i+1])
    
        # This snippet makes sure that the "tip" of the V shaped circuit is also
        # controlled
        cx(gidney_anc[-1], b[-1])
    
        # We now perform the right branch of the V shaped circuit
        
        with invert():
            
            # Call merge to make the QuantumSession enter the inversion environment
            # (doesn't happen automatically with fast_append = 3)
            merge(b[0].qs())
            
            for i in range(len(b)-1):
                
                # Regarding the treatment of the classical values, a similar logic as
                # above is applied
                if i != 0:
                    if ctrl is not None:
                        
                        if "1" == a[i]:
                            cx(ctrl, gidney_anc[i-1])
                        mcx([gidney_anc[i-1], b[i]], gidney_anc[i], method = "gidney", ctrl_state = "11")
                        if "1" == a[i]:
                            cx(ctrl, gidney_anc[i-1])
                    
                    else:
                        mcx([gidney_anc[i-1], b[i]], gidney_anc[i], method = "gidney", ctrl_state = c_inv(a[i]) + "1")
                        
                    cx(gidney_anc[i-1], gidney_anc[i])
                    
                else:
                    if a[i] == "1":
                        if ctrl is None:
                            cx(b[i], gidney_anc[i])
                        else:
                            mcx([ctrl, b[i]], gidney_anc[i], method = "gidney")
        
        # Finally, the CNOT gates on the right side of the V shape are executed.
        
        for i in range(len(a)):
            # Since the values of a (in his paper "i") are classically known we
            # can use a classical condition
            if a[i] == "1":
                
                if c_in is not None and i == 0:
                    continue
                # Here we need to again treat the control qubit, because in the 
                # controlled case, the V shaped part of the circuit might be properly
                # controlled because of the above construction but the right most
                # CNOT gates are not.
                if ctrl is None:
                    x(b[i])
                else:
                    cx(ctrl, b[i])
        
        gidney_anc.delete(verify = False)


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from builtins import id

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np

from qrisp.circuit import QuantumCircuit, fast_append, Instruction, Qubit, QubitDealloc, QubitAlloc, ControlledOperation
from qrisp.permeability.type_checker import is_permeable

def get_perm_dic(gate):
    """
    This function takes an Operation object and returns a dictionary indicating the
    permeability status of that Operation on each of it's qubits.

    Parameters
    ----------
    gate : Operation
        The Operation object to investigate.

    Returns
    -------
    dict
        A dictionary of the type {qubit_index : permeability_type}.

    """
    
    # If the gate is known to be X-permeable return the corresponding dict
    if gate.name in ["x", "rx", "rxx"]:
        return {i : "X" for i in range(gate.num_qubits)}
    
    # This dictionary will contain the result
    res = {}
    
    # The strategy is now to determine Z-permeability with the established function
    for i in range(gate.num_qubits):
        if is_permeable(gate, [i]):
            res[i] = "Z"
        else:
            res[i] = "neutral"
    
    # If the gate is a controlled operation, we can determine based on the permeability of
    # the base operation
    if isinstance(gate, ControlledOperation):
        base_op_perm = get_perm_dic(gate.base_operation)
        for i in range(gate.base_operation.num_qubits):
            res[i+gate.num_qubits-gate.base_operation.num_qubits] = base_op_perm[i]
    
    # The following code documents a general way of determining X-permeability, however
    # it requires the unitary of the gate, which can be expensive.
    # By not doing this, we miss out on permeability information in some situations
    # but have significantly fast uncomputation.
    return res
    
    # To check for X-permeability in general, we wrap the gate in H-gates and 
    # determine Z permeability.
    qc = QuantumCircuit(gate.num_qubits)
    for qb in qc.qubits:    
        qc.h(qb)
    qc.append(gate, qc.qubits)
    for qb in qc.qubits:    
        qc.h(qb)
    
    x_perm_gate = qc.to_gate()
    
    for i in range(gate.num_qubits):
        if res[i] == "Z":
            continue
        elif is_permeable(x_perm_gate, [i]):
            res[i] = "X"

    return res        

# This is the base class of the Nodes that appear in the Permeability DAG
class UnqompNode:
    def __init__(self, name, instr=None):
        self.name = name
        self.uncomputed_node = None
        self.instr = instr
        self.hash = id(self)

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        return self.hash == other.hash

# Node which describes allocations.
# The artificial attribute indicates, whether the Allocation node has been added
# by the DAG construction function or whether it has been in the given QuantumCircuit
class AllocNode(UnqompNode):
    def __init__(self, instr, artificial = False):
        name = "alloc_" + instr.qubits[0].identifier
        UnqompNode.__init__(self, name, instr)
        self.artificial = artificial
    
    def __repr__(self):
        return r"$|0 \rangle$"
    
# Node describing deallocations.
class DeallocNode(UnqompNode):
    def __init__(self, instr):
        name = "dealloc_" + instr.qubits[0].identifier
        UnqompNode.__init__(self, name, instr)
        
    def __repr__(self):
        return r"$\langle 0|$"

# Node describing terminations.
class TerminatorNode(UnqompNode):
    def __init__(self, qubit):
        name = "terminate_" + qubit.identifier
        UnqompNode.__init__(self, name)
        self.qubit = qubit
    
    def __repr__(self):
        return "T"

# Node describing instructions
class InstructionNode(UnqompNode):
    def __init__(self, instr):
        name = instr.op.name
        UnqompNode.__init__(self, name, instr = instr)
        
    def __repr__(self):
        return self.instr.op.name

# This class represents the Permeability Graph
# As a subclass of networkx's DiGraph, it can be processed by 
# many of networkx's algorithms.
class PermeabilityGraph(nx.DiGraph):
    
    def __init__(self, qc, remove_artificials = False):
        
        nx.DiGraph.__init__(self)
        
        # The recent_node_dic attribute is a dictionary of the type
        # {Qubit: UnqompNode}
        # which describes which nodes represent the
        # operations that have been carried ot most 
        # recently on the corresponding Qubit.
        self.recent_node_dic = dag_from_qc(self, qc, remove_artificials = remove_artificials)
    
    # Visualizes the DAG
    def draw(self, layout_seed = None):
        """
        Visualizes the PermeabilityGraph.

        Parameters
        ----------
        layout_seed : int, optional
            A seed for the RNG which can (slightly) influence the visualisation layout if the Graph is arranged in an inconvenient way. The default is None.

        """
        visualize_dag(self, layout_seed)
    
    def to_qc(self, topo_sort = None):
        """
        Turns the PermeabilityGraph into a QuantumCircuit, using a given topological sort algorithm.

        Parameters
        ----------
        topo_sort : callable, optional
            The function to perform the topological sorting. By default, networkx.topological_sort is used.

        Returns
        -------
        res_qc : TYPE
            DESCRIPTION.

        """
        
        if topo_sort is None:
            topo_sort = nx.topological_sort
        res_qc = self.original_qc.clearcopy()
        
        with fast_append():
            for node in topo_sort(self):
                if node.instr:
                    if isinstance(node, AllocNode) and node.artificial:
                        continue
                    res_qc.append(node.instr.op, node.instr.qubits, node.instr.clbits)
                    
        return res_qc
    
    def add_edge(self, in_node, out_node, edge_type, qubits = []):
        """
        Adds an edge (in-place) to the PermeabilityGraph.

        Parameters
        ----------
        in_node : UnqompNode
            The node the edge should start at.
        out_node : UnqompNode
            The node the edge should start at.
        edge_type : str
            A string describing the edge type. Should be either 'X', 'Z', 'neutral' or 'anti_dependency'.
        qubits : list[Qubit], optional
            The list of Qubit the edge connects (only applicable if the edge is not anti-dependency). 
            If the same edge is added multiple times with differing qubits, the qubits are accumulated.
            The default is [].

        """
        
        if in_node == out_node:
            return
        
        if self.has_edge(in_node, out_node):
            nx.DiGraph.get_edge_data(self, in_node, out_node)["qubits"].extend(qubits)
        else:
            nx.DiGraph.add_edge(self, in_node, out_node, edge_type=edge_type, qubits = list(qubits))
                
    def get_target_qubits(self, node):
        """
        Returns the target qubits of a node, i.e. the qubits that have either neutral or X-permeability.

        Parameters
        ----------
        node : UnqompNode
            The node to be investigated.

        Raises
        ------
        Exception
            Tried to retrieve target qubits of TerminatorNode.

        Returns
        -------
        list[Qubit]
            The list of target qubits.

        """
        target_qubits = []
        
        if isinstance(node, AllocNode):
            return list(node.instr.qubits)
        
        in_edges = self.in_edges(node, data = True)
        for (in_node, _, data) in in_edges:
            if data["edge_type"] != "Z":
                target_qubits.extend(data["qubits"])
                
        return target_qubits
    
    def get_control_qubits(self, node):
        """
        Returns the control qubits of a node, i.e. the qubits that have Z-permeability.

        Parameters
        ----------
        node : UnqompNode
            The node to be investigated.

        Raises
        ------
        Exception
            Tried to retrieve control qubits of TerminatorNode.

        Returns
        -------
        list[Qubit]
            The list of control qubits.

        """
        control_qubits = []
        
        if isinstance(node, TerminatorNode):
            raise Exception("Tried to retrieve target qubits of TerminatorNode")
        
        if isinstance(node, AllocNode):
            return list(node.instr.qubits)
        
        in_edges = self.in_edges(node, data = True)
        for (in_node, _, data) in in_edges:
            if data["edge_type"] == "Z":
                control_qubits.extend(data["qubits"])
                
        return control_qubits
    
    def get_control_nodes(self, node):
        """
        Returns the list of control nodes of a node, i.e. the nodes that are connected via an
        outgoing edge with Z-permeability type.

        Parameters
        ----------
        node : UnqompNode
            The node to be investigated.

        Raises
        ------
        Exception
            Tried to retrieve target qubits of TerminatorNode.

        Returns
        -------
        list[UnqompNode]
            The list of control nodes.

        """
        control_nodes = []
        
        if isinstance(node, TerminatorNode):
            raise Exception("Tried to retrieve control qubits of TerminatorNode")
        
        in_edges = self.in_edges(node, data = True)
        for (in_node, _, data) in in_edges:
            if data["edge_type"] == "Z":
                control_nodes.append(in_node)
                
        return control_nodes
    
    def get_target_nodes(self, node):
        """
        Returns the list of control nodes of a node, i.e. the nodes that are connected via an
        outgoing edge with X-permeability or neutral type.

        Parameters
        ----------
        node : UnqompNode
            The node to be investigated.

        Raises
        ------
        Exception
            Tried to retrieve target qubits of TerminatorNode.

        Returns
        -------
        list[UnqompNode]
            The list of target nodes.

        """
        target_nodes = []
        
        if isinstance(node, TerminatorNode):
            raise Exception("Tried to retrieve control qubits of TerminatorNode")
        
        in_edges = self.in_edges(node, data = True)
        for (in_node, _, data) in in_edges:
            if data["edge_type"] != "Z":
                target_nodes.append(in_node)
                
        return target_nodes 
    
    def get_edge_qubits(self, in_node, out_node):
        """
        Returns the list of qubits that a given edge is representing.

        Parameters
        ----------
        in_node : UnqompNode
            The starting node of the edge.
        out_node : UnqompNode
            The end node of the edge.

        Raises
        ------
        Exception
            Tried to retrieve edge qubits from Anti-Dependency edge.

        Returns
        -------
        list[Qubit]
            The list of Qubits.

        """
        return self.get_edge_data(in_node, out_node)["qubits"]
    
    def get_streak_children(self, node, qb):
        successors = self.successors(node)
        return [suc for suc in successors if qb in self.get_edge_data(node, suc)["qubits"]]
    
    def get_edge_type(self, node_out, node_in):
        return self.get_edge_data(node_out, node_in)["edge_type"]
    
    def copy(self):
        self.__class__ = nx.DiGraph
        res = self.copy()
        self.__class__ = PermeabilityGraph
        return res



def dag_from_qc(dag, qc, remove_artificials = False):
    """
    This function receives an (empty) PermeabilityGraph and builds up the corresponding
    nodes/edges according to the given QuantumCircuit.

    Parameters
    ----------
    dag : PermeabilityGraph
        The empty PermeabilityGraph instance to operate on.
    qc : QuantumCircuit
        The QuantumCircuit to turn into the PermeabilityGraph.

    Returns
    -------
    recent_node_dic : dict[Qubit: UnqompNode]
        A dictionary describing which nodes represent the last instruction that was
        applied to a given Qubit.

    """
    
    # This dictionary keeps track of the most recent instruction that was 
    # applied to a given Qubit
    recent_node_dic = {}
    
    # This dictionary tracks what kind of streak each Qubit is currently on.
    streak_dic = {}
    
    # This dictionary tracks a quantity called value layer. This is important for visualisation.
    # The idea behind it is to group each node into a layer. Nodes that are part of a streak 
    # are grouped into the same layer.
    value_layer = {}

    artificial_init_nodes = []
    
    # We iterate through the QuantumCircuit and process each Instruction
    for i in range(len(qc.data)):
        
        # Set alias
        instr = qc.data[i]
        
        
        # We check whether the relevant Qubit already have an allocation node.
        for qb in instr.qubits:
            if qb not in recent_node_dic:
                
                # If the qubit has not been allocated yet but the first instruction being
                # executed is not an allocation, we insert and "artificial" allocation node
                is_artificial = (instr.op.name != "qb_alloc")
                
                # Create the allocation node.
                alloc_node = AllocNode(instr = Instruction(QubitAlloc(), [qb]), 
                                       artificial = is_artificial)
                
                if alloc_node.artificial:
                    artificial_init_nodes.append(alloc_node)
                else:
                    alloc_node.instr = instr
                
                # Faster version of adding a node
                # dag.add_node(node)
                dag._succ[alloc_node] = {}
                dag._pred[alloc_node] = {}
                dag._node[alloc_node] = {}
                
                # Set the value layer attribute to 0
                alloc_node.value_layer = 0
                # alloc_node.qubit = qb
                
                # Set up the entry in the corresponding dictionaries
                recent_node_dic[qb] = alloc_node
                streak_dic[qb] = "neutral"
                value_layer[qb] = 1
                alloc_node.qc_index = 0
                
                

        # If the instruction is an allocation, we already processed it with the
        # code above        
        if instr.op.name == "qb_alloc" and value_layer[qb] == 1:
            continue
        
        # This treats the case of a general Instruction
        elif instr.op.name != "qb_dealloc":
            node = InstructionNode(instr)
            
            # Get the permeability dictionary with the appropriate function
            perm_dic = get_perm_dic(instr.op)
            
        # This treats the case of a deallocation node
        else:
            node = DeallocNode(instr)
            perm_dic = {0 : "neutral"}
        
        # We add the index of the corresponding gate to the node object,
        # because this information can be used for stable topological ordering:
        # The allocation algorithm performs a topological order based on the ancestors
        # of a certain nodes. The algorithm can subsequently use another topological
        # ordering to sort the ancestors. To make the sorting algorithm "stable" ie.
        # it preserves the previous order where possible, we use the indices of the source
        # circuit as sorting index.
        node.qc_index = i

        # Faster version of adding a node
        # dag.add_node(node)
        dag._succ[node] = {}
        dag._pred[node] = {}
        dag._node[node] = {}
        
        # To connect the edges, we iterate over each qubit
        for j in range(len(instr.qubits)):
            
            # Set alias
            qb = instr.qubits[j]

            # Retrieve edge typ
            edge_type = perm_dic[j]

            
            
            # Case streak is continued
            if streak_dic[qb] == edge_type and edge_type != "neutral":
                    dag.add_edge(recent_node_dic[qb], node, edge_type=edge_type, qubits = [qb])
                
            # Case streak is terminated or edge_type is neutral
            else:
                
                # If the streak has more than one member, we insert a terminator node.
                # For that, we get the successors and filter out the edges that don't
                # describe the relevant qubit
                successors = list(dag.successors(recent_node_dic[qb]))
                streak_members = [s for s in successors if qb in dag.get_edge_qubits(recent_node_dic[qb], s)]
                
                # Insert the terminator if there are more than one streak_member to our qubit
                if len(streak_members) > 1:
                    
                    # Create the TerminatorNode
                    terminator = TerminatorNode(qb)
                    # dag.add_node(terminator)
                    dag._succ[terminator] = {}
                    dag._pred[terminator] = {}
                    dag._node[terminator] = {}
                    
                    
                    # We now insert the anti-depedency edges from all streak members
                    # to the terminator node.
                    
                    # To determine the value layer of the Terminator, we find the
                    # successor with the hight value layer and increase it by one.
                    value_layers = []
                    for s in streak_members:
                        
                        # Add anti-depedency edge
                        dag.add_edge(s, terminator, edge_type="anti_dependency", qubits = [qb])
                        value_layers.append(s.value_layer)
                    
                    terminator.value_layer = max(value_layers) + 1
                    value_layer[qb] = terminator.value_layer
                    
                    # Insert the terminator into the recen_node_dic
                    recent_node_dic[qb] = terminator
                
                # If there is only one member of the streak, the new recent_node
                # is that member
                elif len(streak_members) == 1:
                    recent_node_dic[qb] = streak_members[0]
                
                # Add the edge to the dag
                dag.add_edge(recent_node_dic[qb], node, edge_type=edge_type, qubits = [qb])
                
                # Increase the value layer
                value_layer[qb] += 1
                
                # Update the streak dic
                streak_dic[qb] = edge_type
            
            # The value layer of the node is the highest value layer of all qubits
            node.value_layer = max([value_layer[qb] for qb in instr.qubits])
            
    # Save the original_qc
    dag.original_qc = qc
    
    if remove_artificials:
        dag.remove_nodes_from(artificial_init_nodes)
    
    return recent_node_dic

def visualize_dag(G, layout_seed = None):
    """
    Visualizes a PermeabilityGraph, coloring edges and nodes according to their types.

    Parameters
    ----------
    G : PermeabilityGraph
        The PermeabilityGraph to visualize.
    layout_seed : int, optional
        An integer representing a random seed that can (slightly) modify the plot layout. The default is None.

    """

    # The subset_key dictionary groups the nodes according to their value_layers.
    # This dictionary will then be used to generate the plot layout via the nx.multipartite_layout
    subset_key = {}
    
    # Fill the dictionary
    for node in G.nodes():
        if node.value_layer in subset_key:
            subset_key[node.value_layer].append(node)
        else:
            subset_key[node.value_layer] = [node]
    
    # This function is a sort key to sort the nodes in each value layer
    # such that the plot has some resemblance with the QuantumCircuit
    def sort_key(x):
        if isinstance(x, TerminatorNode):
            return G.original_qc.qubits.index(x.qubit)
        else:
            return G.original_qc.qubits.index(x.instr.qubits[-1])
    
    # Sort the entries        
    for key in subset_key.keys():
        subset_key[key].sort(key = sort_key)
        subset_key[key].reverse()
    
    # Determine the layout
    pos = nx.multipartite_layout(G, subset_key = subset_key)
    
    # We now modify the vertical positions of each node a bit.
    # We do this because the default layout can make it hard to distinguish
    # between edges crossing multiple layers.
    
    if layout_seed is not None:
        np.random.seed(layout_seed)
    
    rnd = np.random.random(1)
    for key in subset_key.keys():
        for node in subset_key[key]:
            # Multiply the y coordinates with some number
            pos[node][1] /= len(subset_key[key])**(0.5+rnd)

    
    # Now that positions of the nodes are determined we can start plotting the Graph.
    # For this we first classify into (de)allocation nodes, instruction nodes and terminator nodes.
    allocation_nodes = []
    instruction_nodes = []
    terminator_nodes = []
    labels = {}
    
    for node in G.nodes():
        if isinstance(node, InstructionNode):
            instruction_nodes.append(node)
        elif isinstance(node, TerminatorNode):
            terminator_nodes.append(node)
        elif isinstance(node, DeallocNode):
            allocation_nodes.append(node)
        elif isinstance(node, AllocNode):
            allocation_nodes.append(node)
            
        # The node label is according to the __repr__ function
        labels[node] = node.__repr__()
    
    # Draw the nodes
    nx.draw_networkx_nodes(G, pos, nodelist=allocation_nodes, node_color="tab:blue", node_size=750)
    nx.draw_networkx_nodes(G, pos, nodelist=instruction_nodes, node_color="grey", node_size=750)
    nx.draw_networkx_nodes(G, pos, nodelist=terminator_nodes, node_color="orange", node_size=750)

    # Define node and edge colors
    edge_colors = {'Z': 'green', "neutral": 'grey', "anti_dependency" : "purple", "X" : "red"}
    # Draw edges with colors based on their type
    for edge_type, color in edge_colors.items():
        edges = [(u, v) for u, v, attr in G.edges(data=True) if attr.get('edge_type') == edge_type]
        
        # Draw the edges
        nx.draw_networkx_edges(
            G,
            pos,
            edgelist=edges,
            width=4,
            alpha=0.6,
            edge_color=color,
            arrowstyle = "-|>",
            arrowsize = 15
        )
    
    # Draw labels
    nx.draw_networkx_labels(G, pos, font_size=12, labels = labels, font_color = "white")

    # Deactivate frame
    plt.box(False)
    
    # plt.savefig("mm_dag.png", dpi = 300, bbox_inches = "tight")
    # Show plot
    plt.show()
    




"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.alg_primitives.mcx_algs.circuit_library import gidney_qc, gidney_qc_inv, margolus_qc, ctrl_state_wrap
from qrisp.circuit import QuantumCircuit, Operation

class GidneyLogicalAND(Operation):
    
    def __init__(self, inv = False, ctrl_state = "11"):
        
        definition = QuantumCircuit(3)
        
        if ctrl_state[0] == "0":
            definition.x(definition.qubits[0])
        if ctrl_state[1] == "0":
            definition.x(definition.qubits[1])
        
        definition.append(margolus_qc.to_gate("margolus"), definition.qubits)
        
        if ctrl_state[0] == "0":
            definition.x(definition.qubits[0])
        if ctrl_state[1] == "0":
            definition.x(definition.qubits[1])
        
        name = "uncompiled_gidney_mcx"
        if inv:
            name = name + "_inv"
        Operation.__init__(self, name = name, num_qubits = 3, definition = definition)
        
        self.permeability = {0: True, 1: True, 2: False}
        self.inv = inv
        self.is_qfree = True
        self.ctrl_state = ctrl_state
    
    def inverse(self):
        return GidneyLogicalAND(inv = not self.inv, ctrl_state = self.ctrl_state)
    
    def recompile(self):
        if self.inv:
            compiled_qc = gidney_qc_inv
        else:
            compiled_qc = gidney_qc
        
        name = "compiled_gidney_mcx"
        if self.inv:
            name = name + "_inv"
        
        res = ctrl_state_wrap(compiled_qc, self.ctrl_state).to_op(name)
        res.is_qfree = True
        res.permeability = {0: True, 1: True, 2: False}
        return res



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""
import numpy as np

from qrisp.core import cx, ry

def dicke_state(qv, k):
    """
    Dicke State initialization of a QuantumVariable, based on the deterministic alogrithm in https://arxiv.org/abs/1904.07358. 
    This algorithm creates an equal superposition of Dicke states for a given Hamming weight. The initial input variable has to be within this subspace.

    Parameters
    ----------
    qv : QuantumVariable
        Initial quantum variable to be prepared. Has to be in target subspace.
    k : Int
        The Hamming weight (i.e. number of "ones") for the desired dicke state
        

    Examples
    --------
    We initiate a QuantumVariable in the "0011" state and from this create the Dicke state with Hamming weight 2.

    ::
        
        from qrisp import QuantumVariable, x
        from qrisp.misc.dicke_state import dicke_state
        
        qv = QuantumVariable(4)
        x(qv[2])
        x(qv[3])

        dicke_state(qv, 2)

    """

    n = len(qv)
    for index2 in reversed(range(k+1, n+1)):
        split_cycle_shift(qv, index2, k)

    for index in reversed(range(2,k+1)):
        split_cycle_shift(qv, index, index-1)
    

def split_cycle_shift(qv, highIndex, lowIndex):

    """
    Helper function for Dicke State initialization of a QuantumVariable, based on the deterministic alogrithm in https://arxiv.org/abs/1904.07358. 
    
    Parameters
    ----------
    qv : QuantumVariable
        Initial quantum variable to be prepared. Has to be in target subspace.
    highIndex : Int
        Index for indication of preparation steps, as seen in original algorithm.
    lowIndex : Int
        Index for indication of preparation steps, as seen in original algorithm.
    """

    from qrisp import control

    index_range = [highIndex - i for i in range(lowIndex)]
    for index in index_range:
        param = 2 * np.arccos(np.sqrt((highIndex - index + 1 ) /(highIndex)) )

        if index == highIndex:
            cx(qv[highIndex - 2], qv[highIndex-1]) 
            with control( qv[highIndex-1] ):
                ry(param, qv[highIndex - 2])
            cx(qv[highIndex - 2], qv[highIndex -1])

        else: 
            cx(qv[index -2], qv[highIndex-1]) 
            with control([qv[highIndex -1],qv[index -1]]):
                ry(param, qv[index - 2])
            cx(qv[index -2], qv[highIndex-1]) 


"""
\********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import rz, rzz, x
import numpy as np
import copy
import networkx as nx
from qrisp.algorithms.qiro.qiroproblems.qiro_utils import * 


def create_max_clique_replacement_routine(res, problem_updated):
    """
    Creates a replacement routine for the problem structure, i.e., defines the replacement rules. 
    See the `original paper <https://journals.aps.org/prxquantum/abstract/10.1103/PRXQuantum.5.020327>`_ for a description of the update rules.

    Parameters
    ----------
    res : dict
        Result dictionary of QAOA optimization procedure.
    problem_updated : List
        Updates that happened during the QIRO routine. Consits of the updated problem, a list of Qubits which were found to be positively correlated, i.e. part of the problem solution, 
        and a list Qubits which were found to be negatively correlated, i.e. they contradict solution qubits in accordance with the update rules.  

    Returns
    -------
    new_graph : nx.Graph
        Updated graph for the problem instance.
    solutions : list
        Updated set of solutions to the problem.
    sign : int
        The sign of the correlation.
    exclusions : list
        Updated set of exclusions for the problem.
        
    """
    graph = problem_updated[0]
    solutions = problem_updated[1]
    exclusions =  problem_updated[2]

    orig_edges = [list(item) for item in graph.edges()]
    orig_nodes = list(graph.nodes())

    #get the max_edge and eval the sum and sign
    
    max_item, sign = find_max(orig_nodes, orig_edges , res, solutions)
    if max_item == None:
        return graph, solutions, 0 ,exclusions

    new_graph = copy.deepcopy(graph)

    # we just directly remove vertices from the graph 
    if isinstance(max_item, int):
        if sign < 0:
            border = list(graph.adj[max_item].keys())
            border.append(max_item)
            to_remove = [int(item) for item in graph.nodes() if item not in border]
            new_graph.remove_nodes_from( to_remove)
            solutions.append(max_item)
            exclusions += to_remove

        elif sign > 0:
            #remove item
            new_graph.remove_node(max_item)
            exclusions.append(max_item)

    else:
        if sign > 0:
            #keep the two items in solution and remove all that are not adjacent to both
            
            intersect = list(set(list(graph.adj[max_item[0]].keys())) & set(list(graph.adj[max_item[0]].keys())))
            intersect.append(max_item[0])
            intersect.append(max_item[1])
            to_remove = [int(item) for item in graph.nodes() if item not in intersect]
            new_graph.remove_nodes_from([item for item in graph.nodes() if item not in intersect])
            solutions.append(max_item[0])
            solutions.append(max_item[1])
            exclusions += to_remove

        elif sign < 0:
            #remove all that do not border on either! node
            union = list(graph.adj[max_item[0]].keys())
            union += list(graph.adj[max_item[1]].keys())
            union.append(max_item[0])
            union.append(max_item[1])
            to_remove = [int(item) for item in graph.nodes() if item not in union]
            #to_delete = [item for item in graph.nodes() if item not in union]
            new_graph.remove_nodes_from(to_remove)
            exclusions += to_remove

    return new_graph, solutions, sign, exclusions


def create_max_clique_cost_operator_reduced(problem_updated):
    r"""
    Creates the ``cost_operator`` for the problem instance.
    This operator is adjusted to consider qubits that were found to be a part of the problem solution.

    Parameters
    ----------
    problem_updated : List
        Updates that happened during the QIRO routine. Consits of the updated problem, a list of Qubits which were found to be positively correlated, i.e. part of the problem solution, 
        and a list Qubits which were found to be negatively correlated, i.e. they contradict solution qubits in accordance with the update rules.  
    
    Returns
    -------
    cost_operator : function
        A function receiving a :ref:`QuantumVariable` and a real parameter $\gamma$. This function performs the application of the cost operator.

    """
    problem = problem_updated[0]
    solutions = problem_updated[1]

    G_compl = nx.complement(problem)
    def cost_operator(qv, gamma):
        for pair in list(G_compl.edges()):
            rzz(3*gamma, qv[pair[0]], qv[pair[1]])
            rz(-gamma, qv[pair[0]])
            rz(-gamma, qv[pair[1]])
        for i in problem.nodes():
            if not i in solutions:
                rz(gamma, qv[i])

    return cost_operator




"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

import numpy as np

from qrisp.core import h
from qrisp.alg_primitives import QPE

def quantum_counting(qv, oracle, precision):
    """
    This algorithm estimates the amount of solutions for a given Grover oracle.

    Parameters
    ----------
    qv : QuantumVariable
        The QuantumVariable on which to evaluate.
    oracle : function
        The oracle function.
    precision : int
        The precision to perform the quantum phase estimation with.

    Returns
    -------
    M : float
        An estimate of the amount of solutions.

    Examples
    --------

    We create an oracle, which performs a simple phase flip on the last qubit. ::

        from qrisp import quantum_counting, z, QuantumVariable

        def oracle(qv):
            z(qv[-1])


    We expect half of the state-space of the input to be a solution.

    For 3 qubits, the state space is $2^3 = 8$ dimensional.

    >>> quantum_counting(QuantumVariable(3), oracle, 3)
    3.999999999999999

    For 4 qubits, the state space is $2^4 = 16$ dimensional.

    >>> quantum_counting(QuantumVariable(4), oracle, 3)
    7.999999999999998


    """

    from qrisp import gate_wrap
    from qrisp.grover import diffuser

    @gate_wrap
    def grover_operator(qv):
        oracle(qv)
        diffuser(qv)

    h(qv)
    res = QPE(qv, grover_operator, precision=precision)

    mes_res = res.get_measurement()

    theta = min(list(mes_res.keys())[:1]) * 2 * np.pi

    N = 2**qv.size
    M = N * np.sin(theta / 2) ** 2

    return M


